"""
Access Control Policy logging settings

A few things to note:
- Each ACP has it's own unique LoggingSettings. You can't create a LoggingSettings object and then associate to ACPs

- The LoggingSettings object for an ACP is created automatically when the ACP itself is created.

- Only PUT method is support for changing an existing LoggingSettings for an ACP

- Not all values are present by default. For example severityForPlatformSettingSyslogConfig is only valid if you have
syslogConfigFromPlatformSetting enabled

Not yet supported
- Custom syslog setting for ACP
"""
from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from fmcapi.api_objects import AccessPolicies
from fmcapi.api_objects.helper_functions import true_false_checker
import logging


class LoggingSettings(APIClassTemplate):
    """
    The LoggingSettings object in the FMC
    """

    VALID_JSON_DATA = [
        "enableFileAndMalwareSyslog",
        "fileAndMalwareSyslogSeverity",
        "syslogConfigFromPlatformSetting",
        "severityForPlatformSettingSyslogConfig",
        "type",
        "id",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    VALID_SEVERITY = [
        "ALERT",
        "CRIT",
        "DEBUG",
        "EMERG",
        "ERR",
        "INFO",
        "NOTICE",
        "WARNING",
    ]
    REQUIRED_FOR_PUT = ["id", "type"]
    REQUIRED_FOR_GET = ["acp_id"]

    def __init__(self, fmc, **kwargs):
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for LoggingSettings class.")
        self.fmc = fmc
        self.acp_id = None
        self.type = "LoggingSetting"
        self._syslogConfigFromPlatformSetting = False
        self._severityForPlatformSettingSyslogConfig = "ALERT"
        self._enableFileAndMalwareSyslog = False
        self._fileAndMalwareSyslogSeverity = "ALERT"
        self.parse_kwargs(**kwargs)

    def parse_kwargs(self, **kwargs):
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for AccessRules class.")
        self.get_acp_id(**kwargs)

    def get_acp_id(self, **kwargs):
        if "acp_id" in kwargs:
            self.acp_id = kwargs["acp_id"]
        elif "acp_name" in kwargs:
            acp = AccessPolicies(fmc=self.fmc)
            acp.get(name=kwargs["acp_name"])
            if hasattr(acp, "id"):
                self.acp_id = acp.id
            else:
                logging.warning(
                    f"Access Control Policy {kwargs['name']} not found.  Cannot set up logging for ACP."
                )
        else:
            logging.error("No accessPolicy name or ID was provided.")

    def set_url(self):
        if hasattr(self, "id"):
            self.URL = f"{self.fmc.configuration_url}/policy/accesspolicies/{self.acp_id}/loggingsettings"
        else:
            self.URL = f"{self.fmc.configuration_url}/policy/accesspolicies/{self.acp_id}/loggingsettings?expanded=true"

    def get(self):
        for item in self.REQUIRED_FOR_GET:
            if not hasattr(self, item):
                logging.warning(
                    f"Unable to perform operation due to missing attribute: {item}"
                )
                return
        self.set_url()
        response = self.fmc.send_to_api(method="get", url=self.URL)
        if "items" in response:
            self.set_logging_attributes(response["items"][0])
        self.set_url()

    def set_logging_attributes(self, items):
        for attribute in items:
            if attribute == "metadata":
                continue
            setattr(self, attribute, items[attribute])

    @property
    def enableFileAndMalwareSyslog(self):
        return self._enableFileAndMalwareSyslog

    @enableFileAndMalwareSyslog.setter
    def enableFileAndMalwareSyslog(self, value=False):
        self._enableFileAndMalwareSyslog = true_false_checker(value)

    @property
    def fileAndMalwareSyslogSeverity(self):
        return self._severityForPlatformSettingSyslogConfig

    @fileAndMalwareSyslogSeverity.setter
    def fileAndMalwareSyslogSeverity(self, value="ALERT"):
        if value in self.VALID_SEVERITY:
            self._severityForPlatformSettingSyslogConfig = value
        else:
            self._severityForPlatformSettingSyslogConfig = "ALERT"

    @property
    def syslogConfigFromPlatformSetting(self):
        return self._syslogConfigFromPlatformSetting

    @syslogConfigFromPlatformSetting.setter
    def syslogConfigFromPlatformSetting(self, value=False):
        self._syslogConfigFromPlatformSetting = true_false_checker(value)

    @property
    def severityForPlatformSettingSyslogConfig(self):
        return self._severityForPlatformSettingSyslogConfig

    @severityForPlatformSettingSyslogConfig.setter
    def severityForPlatformSettingSyslogConfig(self, value="ALERT"):
        if value in self.VALID_SEVERITY:
            self._severityForPlatformSettingSyslogConfig = value
        else:
            self._severityForPlatformSettingSyslogConfig = "ALERT"



import logging
import fmcapi
import time
import json

def test__dynamicaccesspolicies(fmc):
    logging.info("Testing DynamicAccessPolicies class.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    dap = fmcapi.DynamicAccessPolicies(fmc=fmc)
    dap.name = namer
    dap.post()
    dap.get()

    dap.authorizationAttributes = [
        {
        "dapRecordName": "TEST_DAP_RECORD",
        "action": "CONTINUE",
        "message": "TEST_MESSAGE",
        "priority": 0
        },
        {
        "dapRecordName": "DfltAccessPolicy",
        "action": "CONTINUE",
        "priority": -1
        }
    ]

    dapXmlConfig_json = '{"dapRecordList":{"dapRecord":{"dapName":{"value":"TEST_DAP_RECORD"},"dapViewsRelation":{"value":"and"},"dapBasicView":{"dapSelection":{"dapPolicy":{"value":"match-any"},"attr":{"name":"aaa.ldap.memberOf","operation":"EQ","value":"test_ad_group"}}}}}}'
    dap.dapXmlConfig_dict = json.loads(dapXmlConfig_json)

    hostscanXmlConfig_json = '{"data":{"config":{"field":{"_type":"dropdown","_name":"logging","_value":"error"}},"multilocation":{"sequence":{"start":{"location":{"_name":"Default"}}}},"location":{"field":[{"_type":"text","_name":"tWindowsCleanerLogoutTitle","_value":"(WebVPN Logout)"},{"_type":"checkbox","_name":"cWindowsCleanerLogoutTitle","_value":"ON"},{"_type":"dropdown","_name":"dWindowsCleanerTimeout","_value":"5"},{"_type":"dropdown","_name":"dCleanerSecureDeletePass","_value":"3"},{"_type":"dropdown","_name":"dTimeout","_value":"5"},{"_type":"dropdown","_name":"dSDSecureDeletePass","_value":"3"},{"_type":"text","_name":"tRestrictedModeUnRestricted","_value":""},{"_type":"text","_name":"tInternetExplorerHomePage","_value":"about:blank"}],"favorite":{"_type":"folder","_value":"Favorites"},"_name":"Default"},"customization":"","hostscan":"","_version":"3.2.1"}}'
    dap.hostscanXmlConfig_dict = json.loads(hostscanXmlConfig_json)

    dap.put()
    dap.delete()

    logging.info("Testing DynamicAccessPolicies class done.\n")


"""DNS Server Groups Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class DNSServerGroups(APIClassTemplate):
    """The DNSServerGroups Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "retries",
        "timeout",
        "dnsservers",
        "defaultdomain",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/object/dnsservergroups"
    REQUIRED_FOR_POST = ["name", "timeout"]
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""

    def __init__(self, fmc, **kwargs):
        """
        Initialize DNSServerGroups object.

        Set self.type to "DNSServerGroupObject" and parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for DNSServerGroups class.")
        self.parse_kwargs(**kwargs)
        self.type = "DNSServerGroupObject"

    def servers(self, action, name_servers):
        """
        Associate DNS Servers.

        :param action: (str) 'add', 'remove', or 'clear'
        :param name_servers: (str)  Name of DNS server.
        """
        logging.debug("In servers() for DNSServerGroups class.")
        if action == "add":
            for name_server in name_servers:
                if "dnsservers" in self.__dict__:
                    self.dnsservers.append({"name-server": name_server})
                else:
                    self.dnsservers = [{"name-server": name_server}]
                logging.info(
                    f'Name-server "{name_server}" added to this DNSServerGroups object.'
                )
        elif action == "remove":
            if "dnsservers" in self.__dict__:
                for name_server in name_servers:
                    self.dnsservers = list(
                        filter(
                            lambda i: i["name-server"] != name_server, self.dnsservers
                        )
                    )
            else:
                logging.warning(
                    "DNSServerGroups has no members.  Cannot remove name-server."
                )
        elif action == "clear":
            if "dnsservers" in self.__dict__:
                del self.dnsservers
                logging.info(
                    "All name-servers removed from this DNSServerGroups object."
                )



"""FQDNs Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class FQDNS(APIClassTemplate):
    """The FQDNS Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "description",
        "type",
        "overridableTargetId",
        "value",
        "dnsResolution",
        "overrides",
        "overridable",
    ]
    VALID_GET_FILTERS = [
        "unusedOnly",
        "nameOrValue",
    ]  # unusedOnly:Bool, nameOrValue:String
    VALID_FOR_KWARGS = VALID_JSON_DATA + VALID_GET_FILTERS + []
    URL_SUFFIX = "/object/fqdns"
    VALID_FOR_DNS_RESOLUTION = ["IPV4_ONLY", "IPV6_ONLY", "IPV4_AND_IPV6"]
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""
    FIRST_SUPPORTED_FMC_VERSION = "6.3.0"

    def __init__(self, fmc, **kwargs):
        """
        Initialize FQDNS object.

        Set self.type to "FQDNS" and parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for FQDNS class.")
        self.parse_kwargs(**kwargs)
        self.type = "FQDN"
        if self.fmc.serverVersion < self.FIRST_SUPPORTED_FMC_VERSION:
            logging.warning(
                f"The FQDNS API feature was released in version {self.FIRST_SUPPORTED_FMC_VERSION}.  "
                f"Your FMC version is {self.fmc.serverVersion}.  Upgrade to use this feature."
            )

    def format_data(self):
        """
        Gather all the data in preparation for sending to API in JSON format.

        :param filter_query: (str) 'all' or 'kwargs'
        :return: (dict) json_data
        """
        json_data = super().format_data()
        logging.debug("In format_data() for FQDNS class.")
        if "dnsResolution" in self.__dict__:
            if self.dnsResolution in self.VALID_FOR_DNS_RESOLUTION:
                json_data["dnsResolution"] = self.dnsResolution
            else:
                logging.warning(f"dnsResolution {self.dnsResolution} not a valid type.")
        return json_data

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for FQDNS class.")
        if "dnsResolution" in kwargs:
            if kwargs["dnsResolution"] in self.VALID_FOR_DNS_RESOLUTION:
                self.dnsResolution = kwargs["dnsResolution"]
            else:
                logging.warning(
                    f"dnsResolution {kwargs['dnsResolution']} not a valid type."
                )



import logging
import fmcapi


def test__fmc_version(fmc):
    logging.info(
        "Testing ServerVersion() method.  Getting version information information from FMC."
    )

    version_info = fmcapi.ServerVersion(fmc=fmc)
    logging.info(version_info.get())
    logging.info("Testing ServerVersion() done.\n\n")



import logging
import fmcapi
import time

def test__grouppolicies(fmc):
    logging.info("Testing GroupPolicies class.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    group_policy = fmcapi.GroupPolicies(fmc=fmc)
    group_policy.name = namer
    # Posting a group policy with just a name will create a blank group policy with whatever defaults the GUI uses at the moment
    group_policy.post()

    group_policy.get()

    # Change number of simultaneous ravpn sessions per user as PUT example
    # Note: simultaneousLoginPerUser = 0 via the api fails to deploy
    # because of this active bug - CSCwi89739 - even though 0 is a valid value
    # and can be successfully deployed in the GUI
    sessionSettings = group_policy.advancedSettings.get('sessionSettings')
    sessionSettings['simultaneousLoginPerUser'] = 1
    group_policy.advancedSettings['sessionSettings'] = sessionSettings
    group_policy.put()

    group_policy.get()
    if group_policy.advancedSettings.get('sessionSettings').get('simultaneousLoginPerUser') == 1:
        group_policy.delete()

    logging.info("Testing GroupPolicies class done.\n")


from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .dynamicobjects import DynamicObject
import logging


class DynamicObjectMappings(APIClassTemplate):
    """The Dynamic Object in the FMC."""

    VALID_JSON_DATA = ["add", "remove", "mappings", "type", "id", "name"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/object/dynamicobjectmappings"
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""

    REQUIRED_FOR_POST = []
    obj_id = ""

    def __init__(self, fmc, **kwargs):
        """
        Initialize Dynamic Object.

        Set self.type to "DynamicObject" and parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for Dynamic Object class.")
        self.parse_kwargs(**kwargs)
        if "id" in kwargs:
            self.obj_id = kwargs["id"]

    def handle_mappings(self, action, value, name):
        """
        Associate mappings to dynamic object.

        :param action: (str) 'add', 'remove', or 'clear'
        :param dynamic_objects: (DynamicObject) Dynamic object to be mapped.
        """
        logging.debug("In mappings for dynamic object")

        dynamic_object = DynamicObject(fmc=self.fmc, name=name)
        response = dynamic_object.get()
        if response:
            new_obj = {
                "dynamicObject": {
                    "name": response["name"],
                    "id": response["id"],
                    "type": response["type"],
                },
                "mappings": value,
            }
            if action == "add":
                if "add" in self.__dict__:
                    self.add.append(new_obj)

                else:
                    self.add = [new_obj]

                logging.info(f"Adding mappings to Dynamic Object.")

            elif action == "remove":
                if "remove" in self.__dict__:
                    self.remove.append(new_obj)

                else:
                    self.remove = [new_obj]

                logging.info(f"Remove mappings to Dynamic Object.")

        else:
            logging.warning(
                f'Dynamic Object "{name}" is not found in FMC.  Cannot add mappings.'
            )

    def get(self):
        self.URL_SUFFIX = f"/object/dynamicobjects/{str(self.obj_id)}/mappings"
        self.__init__(self.fmc)
        return self._get()

    def put(self):
        self.URL_SUFFIX = f"/object/dynamicobjects/{str(self.obj_id)}/mappings"
        super().__init__(self.fmc)
        return super().put()

    def delete(self):
        self.URL_SUFFIX = f"/object/dynamicobjects/{str(self.obj_id)}/mappings"
        super().__init__(self.fmc)
        return super().delete()

    def _get(self, **kwargs):
        """
        To get mappings url is GET /api/fmc_config/v1/domain/{domainUUID}/object/dynamicobjects/{objectId}/mappings
        """
        logging.debug("In get() for APIClassTemplate class.")
        self.parse_kwargs(**kwargs)
        if self.fmc.serverVersion < self.FIRST_SUPPORTED_FMC_VERSION:
            logging.error(
                f"Your FMC version, {self.fmc.serverVersion} does not support GET of this feature."
            )
            return {"items": []}
        if self.valid_for_get():
            if "id" in self.__dict__ or "targetId" in self.__dict__:
                if "targetId" in self.__dict__:
                    url = f"{self.URL}"
                    if "backupVersion" in self.__dict__:
                        url += f"?backupVersion={self.backupVersion}"
                else:
                    url = f"{self.URL}"
                if self.dry_run:
                    logging.info(
                        "Dry Run enabled.  Not actually sending to FMC.  Here is what would have been sent:"
                    )
                    logging.info("\tMethod = GET")
                    logging.info(f"\tURL = {self.URL}")
                    return False
                response = self.fmc.send_to_api(method="get", url=url)
                try:
                    self.parse_kwargs(**response)
                except TypeError as e:
                    logging.error(
                        f"Response from FMC GET with 'id', {self.id}, returned none."
                        f"That 'id' probably was deleted.  Error: {e}."
                    )
                if "name" in self.__dict__:
                    logging.info(
                        f'GET success. Object with name: "{self.name}" and id: "{self.id}" fetched from FMC.'
                    )
                elif "targetId" in self.__dict__:
                    if "backupVersion" in self.__dict__:
                        logging.info(
                            f'GET success. Object with targetId: "{self.targetId}" backupVersion: "{self.backupVersion}" fetched from FMC.'
                        )
                    logging.info(
                        f'GET success. Object with targetId: "{self.targetId}" fetched from FMC.'
                    )
                else:
                    logging.info(
                        f'GET success. Object with id: "{self.id}" fetched from FMC.'
                    )
            elif "name" in self.__dict__:
                if self.FILTER_BY_NAME:
                    url = f"{self.URL}?name={self.name}&expanded=true"
                else:
                    url = f"{self.URL}?expanded=true"
                    if "limit" in self.__dict__:
                        url = f"{url}&limit={self.limit}"
                    if "offset" in self.__dict__:
                        url = f"{url}&offset={self.offset}"
                response = self.fmc.send_to_api(method="get", url=url)
                if "items" not in response:
                    response["items"] = []
                for item in response["items"]:
                    if "name" in item:
                        if item["name"] == self.name:
                            self.id = item["id"]
                            self.parse_kwargs(**item)
                            logging.info(
                                f'GET success. Object with name: "{self.name}" and id: "{self.id}" '
                                f"fetched from FMC."
                            )
                            return item
                    else:
                        logging.warning(
                            f'No "name" attribute associated with this item to check against {self.name}.'
                        )
                if "id" not in self.__dict__:
                    logging.warning(f"\tGET query for {self.name} is not found.")
                    logging.debug(
                        f"\tGET query for {self.name} is not found.\n\t\tResponse: {json.dumps(response)}"
                    )
            elif len(self.get_filters) > 0:
                url_filter = ""
                for key, value in self.get_filters.items():
                    # Filter value must not be empty otherwise will result in a 400 response
                    if value != "":
                        url_filter += f"{key}%3A{value};"
                    else:
                        logging.warning(
                            f"Terminating GET - {self.URL}?expanded={self.expanded}&filter={url_filter}"
                        )
                        logging.warning(f"{key} MUST have a non empty value")
                        return False
                url = f"{self.URL}?expanded={self.expanded}&filter={url_filter}"
                if self.dry_run:
                    logging.info(
                        "Dry Run enabled.  Not actually sending to FMC.  Here is what would have been sent:"
                    )
                    logging.info("\tMethod = GET")
                    logging.info(f"\tURL = {url}")
                    return False
                response = self.fmc.send_to_api(method="get", url=url)
                if "items" not in response:
                    logging.info(
                        f"GET success. No Objects were found with query filter: {self.get_filters}"
                    )
                    return response
                else:
                    response_count = response.get("paging").get("count")
                    logging.info(
                        f"GET success. {response_count} items found that match query filter: {self.get_filters}"
                    )
                    return response
            else:
                logging.debug(
                    "GET query for object with no name or id set.  "
                    "Returning full list of these object types instead."
                )
                url_suffix_start = "?"
                if url_suffix_start in self.URL:
                    url_suffix_start = "&"
                url = f"{self.URL}{url_suffix_start}expanded=true&limit={self.limit}"
                if self.dry_run:
                    logging.info(
                        "Dry Run enabled.  Not actually sending to FMC.  Here is what would have been sent:"
                    )
                    logging.info("\tMethod = GET")
                    logging.info(f"\tURL = {self.URL}")
                    return False
                response = self.fmc.send_to_api(method="get", url=url)
            if "items" not in response:
                response["items"] = []
            return response
        else:
            logging.warning(
                "get() method failed due to failure to pass valid_for_get() test."
            )
            return False



"""Interface Objects Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class InterfaceObjects(APIClassTemplate):
    """The InterfaceObjects Object in the FMC."""

    URL_SUFFIX = "/object/interfaceobjects"
    FILTER_BY_NAME = True

    def __init__(self, fmc, **kwargs):
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for InterfaceObjects class.")
        self.parse_kwargs(**kwargs)

    def post(self):
        """POST method for API for InterfaceObjects not supported."""
        logging.info("POST method for API for InterfaceObjects not supported.")
        pass

    def put(self):
        """PUT method for API for InterfaceObjects not supported."""
        logging.info("PUT method for API for InterfaceObjects not supported.")
        pass

    def delete(self):
        """DELETE method for API for InterfaceObjects not supported."""
        logging.info("DELETE method for API for InterfaceObjects not supported.")
        pass



import logging
import fmcapi
import time

def test__extended_acls(fmc):
    logging.info("Testing ExtendedAccessList class.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    ext_acl = fmcapi.ExtendedAccessList(fmc=fmc)
    ext_acl.name = namer
    ext_acl.entries = []

    ace = fmcapi.ExtendedAccessListAce()
    ace.action = "DENY"
    ace.destinationNetworksLiterals = [
        {
            "type" : "Host",
            "value": "1.1.1.1"
        }
    ]

    ext_acl.entries.append(ace.build_ace())
    ext_acl.post()
    logging.debug(ext_acl.get())

    ace.destinationNetworksLiterals = [
        {
            "type" : "Host",
            "value": "2.2.2.2"
        }
    ]

    ext_acl.entries.append(ace.build_ace())
    logging.debug(ext_acl.entries)
    ext_acl.put()
    logging.debug(ext_acl.get())
    ext_acl.delete()
    del ext_acl
    logging.info("Testing ExtendedAccessList class done.\n")



from fmcapi.api_objects import LoggingSettings, AccessPolicies
import random
import string


def test_logging_settings(fmc):

    suffix = "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(5)
    )

    acp_rule = AccessPolicies(fmc=fmc, name=f"acp-log-test-{suffix}")
    acp_rule.post()

    acp_logging = LoggingSettings(fmc=fmc, acp_name=f"acp-log-test-{suffix}")
    acp_logging.get()

    acp_logging.syslogConfigFromPlatformSetting = True
    acp_logging.severityForPlatformSettingSyslogConfig = "ERR"
    acp_logging.put()

    acp_logging_test = LoggingSettings(fmc=fmc, acp_name=f"acp-log-test-{suffix}")
    acp_logging_test.get()
    if not acp_logging_test.syslogConfigFromPlatformSetting:
        print(
            f"Test Failed: Expected syslogConfigFromPlatformSetting = true, Actual = false"
        )
    if not acp_logging_test.severityForPlatformSettingSyslogConfig == "ERR":
        print(
            f"Test Failed: Expected severityForPlatformSettingSyslogConfig = 'ERR', "
            f"Actual = '{acp_logging.severityForPlatformSettingSyslogConfig}'"
        )

    acp_logging.enableFileAndMalwareSyslog = True
    acp_logging.fileAndMalwareSyslogSeverity = "CRIT"
    acp_logging.put()

    acp_logging_test = LoggingSettings(fmc=fmc, acp_name=f"acp-log-test-{suffix}")
    acp_logging_test.get()
    if not acp_logging_test.enableFileAndMalwareSyslog:
        print(
            f"Test Failed: Expected enableFileAndMalwareSyslog = true, Actual = false"
        )
    if not acp_logging_test.fileAndMalwareSyslogSeverity == "CRIT":
        print(
            f"Test Failed: Expected fileAndMalwareSyslogSeverity = 'CRIT', "
            f"Actual = {acp_logging_test.fileAndMalwareSyslogSeverity}"
        )

    acp_rule.delete()



"""IKE Settings Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .ftds2svpns import FTDS2SVPNs
from fmcapi.api_objects.object_services.ikev1policies import IKEv1Policies
from fmcapi.api_objects.object_services.ikev2policies import IKEv2Policies
from fmcapi.api_objects.object_services.certenrollments import CertEnrollments
import logging


class IKESettings(APIClassTemplate):
    """The IKESettings Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "ikeV1Settings",
        "ikeV2Settings",
        "description",
        "version",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    FIRST_SUPPORTED_FMC_VERSION = "6.3"
    PREFIX_URL = "/policy/ftds2svpns"
    REQUIRED_FOR_POST = ["vpn_id"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize IKESettings object.

        Set self.type to "IKESetting", parse the kwargs, and set up the self.URL.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for IKESettings class.")
        self.parse_kwargs(**kwargs)
        self.type = "IKESetting"

    def vpn_policy(self, pol_name):
        """
        Associate VPN Policy.

        :param pol_name: (str) Name of VPN Policy.
        :return: None
        """
        logging.debug("In vpn_policy() for IKESettings class.")
        ftd_s2s = FTDS2SVPNs(fmc=self.fmc)
        ftd_s2s.get(name=pol_name)
        if "id" in ftd_s2s.__dict__:
            self.vpn_id = ftd_s2s.id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.vpn_id}/ikesettings"
            self.vpn_added_to_url = True
        else:
            logging.warning(
                f'FTD S2S VPN Policy "{pol_name}" not found.  Cannot set up IKESettings for FTDS2SVPNs Policy.'
            )

    def ike_policy(self, pol_name, version=1):
        """
        Associate IKE Policy.

        :param pol_name: (str) Name of IKE Policy.
        :param version: (int) IKE version.
        :return: None
        """
        # ikev1 and ikv2 policy names can overlap
        logging.debug("In ike_policy() for IKESettings class.")
        pol1 = None
        if version == 1:
            pol1 = IKEv1Policies(fmc=self.fmc)
        elif version == 2:
            pol1 = IKEv2Policies(fmc=self.fmc)
        else:
            logging.warning("Invalid version type specified.  Must be between 1-2")
        pol1.get(name=pol_name)

        if "id" in pol1.__dict__ and version == 1:
            self.ikeV1Settings["policy"] = {
                "id": pol1.id,
                "name": pol1.name,
                "type": pol1.type,
            }
        elif "id" in pol1.__dict__ and version == 2:
            self.ikeV2Settings["policy"] = {
                "id": pol1.id,
                "name": pol1.name,
                "type": pol1.type,
            }
        else:
            logging.warning(
                f'IKEv"{version}"Policy "{pol_name}" not found.  Cannot set up IKESettings Policy.'
            )

    def certificate(self, cert_name, version=1):
        """
        Associate a Certificate.

        :param cert_name: (str) Name of certificate.
        :param version: (int) Version of certificate.
        """
        logging.debug("In certificate() for IKESettings class.")
        cert1 = CertEnrollments(fmc=self.fmc)
        cert1.get(name=cert_name)
        if "id" in cert1.__dict__ and version == 1:
            self.ikeV1Settings["authenticationType"] = "CERTIFICATE"
            self.ikeV1Settings["certificateAuth"] = {
                "name": cert1.name,
                "id": cert1.id,
                "type": cert1.type,
            }
        elif "id" in cert1.__dict__ and version == 2:
            self.ikeV2Settings["authenticationType"] = "CERTIFICATE"
            self.ikeV2Settings["certificateAuth"] = {
                "name": cert1.name,
                "id": cert1.id,
                "type": cert1.type,
            }
        else:
            logging.warning(
                f'Certificate "{cert_name}" not found.  Cannot set up IKESettings Policy.'
            )



"""Search Classes."""

import logging
from .globalsearch import GlobalSearch
from .object import Object
from .policy import Policy

logging.debug("In the object_services __init__.py file.")

__all__ = [
    "GlobalSearch",
    "Object",
    "Policy",
]



"""IKE v2 Policies Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class IKEv2Policies(APIClassTemplate):
    """The IKEv2Policies Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "priority",
        "diffieHellmanGroups",
        "integrityAlgorithms",
        "prfIntegrityAlgorithms",
        "encryptionAlgorithms",
        "lifetimeInSeconds",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/object/ikev2policies"
    REQUIRED_FOR_POST = [
        "name",
        "integrityAlgorithms",
        "prfIntegrityAlgorithms",
        "encryptionAlgorithms",
        "diffieHellmanGroups",
    ]
    VALID_FOR_ENCRYPTION = [
        "DES",
        "3DES",
        "AES",
        "AES-192",
        "AES-256",
        "NULL",
        "AES-GCM",
        "AES-GCM-192",
        "AES-GCM-256",
    ]
    VALID_FOR_INTEGRITY = ["NULL", "MD5", "SHA", "SHA-256", "SHA-384", "SHA-512"]
    VALID_FOR_PRF_INTEGRITY = ["MD5", "SHA", "SHA-256", "SHA-384", "SHA-512"]
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""
    FIRST_SUPPORTED_FMC_VERSION = "6.3"

    def __init__(self, fmc, **kwargs):
        """
        Initialize IKEv2Policies object.

        Set self.type to "IKEv2PolicyObject" and parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for IKEv2Policies class.")
        self.parse_kwargs(**kwargs)
        self.type = "Ikev2PolicyObject"

    def encryption(self, action, algorithms=[]):
        """
        Associate encryption.

        :param action: (str) 'add', 'remove', or 'clear'
        :param algorithms: (list) List of algorithms.
        :return: None
        """
        logging.debug("In encryption() for IKEv2Policies class.")
        if action == "add":
            for algorithm in algorithms:
                if "encryptionAlgorithms" in self.__dict__:
                    if algorithm in self.encryptionAlgorithms:
                        logging.warning(
                            f'encryptionAlgorithms "{algorithm}" already exists.'
                        )
                    elif algorithm in self.VALID_FOR_ENCRYPTION:
                        self.encryptionAlgorithms.append(algorithm)
                    else:
                        logging.warning(
                            f'encryptionAlgorithms "{algorithm}" not a valid type.'
                        )
                else:
                    self.encryptionAlgorithms = [algorithm]
        elif action == "remove":
            if "encryptionAlgorithms" in self.__dict__:
                for algorithm in algorithms:
                    self.encryptionAlgorithms = list(
                        filter(lambda i: i != algorithm, self.encryptionAlgorithms)
                    )
            else:
                logging.warning(
                    "IKEv2Policies has no members.  Cannot remove encryptionAlgorithms."
                )
        elif action == "clear":
            if "encryptionAlgorithms" in self.__dict__:
                del self.encryptionAlgorithms
                logging.info(
                    "All encryptionAlgorithms removed from this IKEv2Policies object."
                )

    def hash(self, action, algorithms=[]):
        """
        Associate hash.

        :param action: (str) 'add', 'remove', or 'clear'
        :param algorithms: (list) List of algorithms.
        :return: None
        """
        logging.debug("In hash() for IKEv2Policies class.")
        if action == "add":
            for algorithm in algorithms:
                if "integrityAlgorithms" in self.__dict__:
                    if algorithm in self.integrityAlgorithms:
                        logging.warning(
                            f'integrityAlgorithms "{algorithm}" already exists.'
                        )
                    elif algorithm in self.VALID_FOR_INTEGRITY:
                        self.integrityAlgorithms.append(algorithm)
                    else:
                        logging.warning(
                            f'integrityAlgorithms "{algorithm}" not a valid type.'
                        )
                else:
                    if algorithm in self.VALID_FOR_INTEGRITY:
                        self.integrityAlgorithms = [algorithm]
                    else:
                        logging.warning(
                            f'integrityAlgorithms "{algorithm}" not a valid type.'
                        )
        elif action == "remove":
            if "integrityAlgorithms" in self.__dict__:
                for algorithm in algorithms:
                    self.integrityAlgorithms = list(
                        filter(lambda i: i != algorithm, self.integrityAlgorithms)
                    )
            else:
                logging.warning(
                    "IKEv2Policies has no members.  Cannot remove integrityAlgorithms."
                )
        elif action == "clear":
            if "integrityAlgorithms" in self.__dict__:
                del self.integrityAlgorithms
                logging.info(
                    "All integrityAlgorithms removed from this IKEv2Policies object."
                )

    def prf_hash(self, action, algorithms=[]):
        """
        Associate prf_hash.

        :param action: (str) 'add', 'remove', or 'clear'
        :param algorithms: (list) List of algorithms.
        :return: None
        """
        logging.debug("In prf_hash() for IKEv2Policies class.")
        if action == "add":
            for algorithm in algorithms:
                if "prfIntegrityAlgorithms" in self.__dict__:
                    if algorithm in self.prfIntegrityAlgorithms:
                        logging.warning(
                            f'prfIntegrityAlgorithms "{algorithm}" already exists.'
                        )
                    elif algorithm in self.VALID_FOR_PRF_INTEGRITY:
                        self.prfIntegrityAlgorithms.append(algorithm)
                    else:
                        logging.warning(
                            f'prfIntegrityAlgorithms "{algorithm}" not a valid type.'
                        )
                else:
                    if algorithm in self.VALID_FOR_PRF_INTEGRITY:
                        self.prfIntegrityAlgorithms = [algorithm]
                    else:
                        logging.warning(
                            f'prfIntegrityAlgorithms "{algorithm}" not a valid type.'
                        )
        elif action == "remove":
            if "prfIntegrityAlgorithms" in self.__dict__:
                for algorithm in algorithms:
                    self.prfIntegrityAlgorithms = list(
                        filter(lambda i: i != algorithm, self.prfIntegrityAlgorithms)
                    )
            else:
                logging.warning(
                    "IKEv2Policies has no members.  Cannot remove prfIntegrityAlgorithms."
                )
        elif action == "clear":
            if "prfIntegrityAlgorithms" in self.__dict__:
                del self.prfIntegrityAlgorithms
                logging.info(
                    "All prfIntegrityAlgorithms removed from this IKEv2Policies object."
                )



"""Device HA Pair Services Classes."""

import logging
from .ftddevicehapairs import FTDDeviceHAPairs
from .failoverinterfacemacaddressconfigs import FailoverInterfaceMACAddressConfigs
from .monitoredinterfaces import MonitoredInterfaces

logging.debug("In the device_ha_pair_services __init__.py file.")

__all__ = [
    "FTDDeviceHAPairs",
    "FailoverInterfaceMACAddressConfigs",
    "MonitoredInterfaces",
]



import logging
import fmcapi


def test__application_type(fmc):
    logging.info("Testing ApplicationType class.")

    obj1 = fmcapi.ApplicationTypes(fmc=fmc)
    logging.info("All ApplicationType -- >")
    result = obj1.get()
    logging.info(result)
    logging.info(f"Total items: {len(result['items'])}")

    del obj1
    obj1 = fmcapi.ApplicationTypes(fmc=fmc, name="Server")
    logging.info("One ApplicationType -- >")
    logging.info(obj1.get())

    logging.info("Testing ApplicationType class done.\n")



"""Misc methods/functions that are used by the fmcapi package's modules."""

import re
import ipaddress
import json
import logging

logging.debug(f"In the {__name__} module.")


def true_false_checker(value):
    if type(value) is str:
        value = value.lower()
        if value == "true":
            return True
        elif value == "false":
            return False
    elif type(value) is int:
        if value == 1:
            logging.warning(
                f"Value, {value}, should be True or False.  Assuming you meant True."
            )
            return True
        elif value == 0:
            logging.warning(
                f"Value, {value}, should be True or False.  Assuming you meant False."
            )
            return False
    elif type(value) is bool:
        return value

    logging.warning(f"Invalid value: '{value}'. Should be True or False")
    return value


def syntax_correcter(value, permitted_syntax="""[.\w\d_\-]""", replacer="_"):
    """
    Check 'value' for invalid characters (identified by 'permitted_syntax') and replace them with 'replacer'.

    :param value:  (str) String to be checked.
    :param permitted_syntax: (str) regex of allowed characters.
    :param replacer: (str) character used to replace invalid characters.
    :return: (str) Modified string with "updated" characters.
    """
    logging.debug("In syntax_correcter() helper_function.")
    new_value = ""
    for char in range(0, len(value)):
        if not re.match(permitted_syntax, value[char]):
            new_value += replacer
        else:
            new_value += value[char]
    return new_value


def get_networkaddress_type(value):
    """
    Check to see whether 'value' is a host, range, or network.

    :param value: (str) x.x.x.x, x.x.x.x/xx, or x.x.x.x-x.x.x.x
    :return: (str) 'host', 'network', or 'range'
    """
    logging.debug("In get_networkaddress_type() helper_function.")
    if "/" in value:
        ip, bitmask = value.split("/")
        if ip == "32" or bitmask == "128":
            return "host"
        else:
            return "network"
    else:
        if "-" in value:
            return "range"
        else:
            return "host"


def is_ip(ip):
    """
    Check whether the provided string is an IP address.

    :param ip: (str) x.x.x.x
    :return: (boolean)
    """
    logging.debug("In is_ip() helper_function.")
    try:
        ipaddress.ip_address(ip)
    except ValueError as err:
        logging.error(err)
        return False
    return True


def is_ip_network(ip):
    """
    Check whether provided string is a valid network address.

    See if the provided IP/SM is the "network address" of the subnet.

    :param ip: (str) x.x.x.x/xx
    :return: (boolean)
    """
    logging.debug("In is_ip_network() helper_function.")
    try:
        ipaddress.ip_network(ip)
    except ValueError as err:
        logging.error(err)
        return False
    return True


def validate_ip_bitmask_range(value="", value_type=""):
    """
    We need to check the provided IP address (or range of addresses) and make sure the IPs are valid.

    :param value: (str) x.x.x.x, x.x.x.x/xx, or x.x.x.x-x.x.x.x
    :param value_type: (str) 'host', 'network', or 'range'
    :return: (dict) {value=value_fixed, valid=boolean}
    """
    logging.debug("In validate_ip_bitmask_range() helper_function.")
    return_dict = {"value": value, "valid": False}
    if value_type == "range":
        for ip in value.split("-"):
            if is_ip(ip):
                return_dict["valid"] = True
    elif value_type == "host" or value_type == "network":
        if is_ip_network(value):
            return_dict["valid"] = True
    return return_dict["valid"]


def mocked_requests_get(**kwargs):
    """
    Use to "mock up" a response from using the "requests" library to avoid actually using the "requests" library.

    :param kwargs:
    :return: (boolean)
    """
    logging.debug("In mocked_requests_get() helper_function.")

    class MockResponse:
        def __init__(self, **kwargs):
            logging.debug("In MockResponse __init__ method.")
            self.text = json.dumps(kwargs["text"])
            self.status_code = kwargs["status_code"]

        def close(self):
            logging.debug("In MockResponse close method.")
            return True

    return MockResponse(**kwargs)


def validate_vlans(start_vlan, end_vlan=""):
    """
    Validate that the start_vlan and end_vlan numbers are in 1 - 4094 range.  If not, then return 1, 4094.

    :param start_vlan: (int) Lower VLAN number in range.
    :param end_vlan: (int) Upper VLAN number in range.
    :return: (int) start_vlan, (int) end_vlan)
    """
    logging.debug("In validate_vlans() helper_function.")
    if end_vlan == "":
        end_vlan = start_vlan
    if int(end_vlan) < int(start_vlan):
        start_vlan, end_vlan = end_vlan, start_vlan
    if 0 < int(start_vlan) < 4095 and 0 < int(end_vlan) < 4095:
        return start_vlan, end_vlan
    else:
        return 1, 4094



"""IKE v1 IPSec Proposals Object Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class IKEv1IpsecProposals(APIClassTemplate):
    """The IKEv1IpsecProposals Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "type", "espEncryption", "espHash"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/object/ikev1ipsecproposals"
    REQUIRED_FOR_POST = ["name", "espEncryption", "espHash"]
    VALID_FOR_ENCRYPTION = ["DES", "3DES", "AES-128", "AES-192", "AES-256", "ESP-NULL"]
    VALID_FOR_HASH = ["NONE", "MD5", "SHA"]
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""
    FIRST_SUPPORTED_FMC_VERSION = "6.3"

    def __init__(self, fmc, **kwargs):
        """
        Initialize IKEv1IpsecProposals object.

        Set self.type to "IKEv1IpsecProposal" and  parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for IKEv1IpsecProposals class.")
        self.parse_kwargs(**kwargs)
        self.type = "IKEv1IPsecProposal"



"""
Establish and manage connection to FMC.

This module (fmc.py) is designed to provide a "toolbox" of tools for interacting with the Cisco FMC API.
The "toolbox" is the FMC class and the "tools" are its methods.  Note: There exists a "Quick Start Guide" for the Cisco
FMC API too.  Just Google for it as it gets updated with each release of code.
"""

import datetime
import requests
import time
import json
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import logging
from logging.handlers import RotatingFileHandler
from .api_objects import ServerVersion
from .api_objects import DeploymentRequests
from sys import exit

# Disable annoying HTTP warnings
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

""""
The 'requests' package is very chatty on the INFO logging level.  Change its logging threshold sent to logger to
something greater than INFO (i.e. not INFO or DEBUG) will cause it to not log its INFO and DEBUG messages to the
default logger.  This reduces the size of our log files.
"""
logging.getLogger("requests").setLevel(logging.WARNING)


class FMC(object):
    """Establish and maintain connection to Firepower Management Center."""

    logging.debug("In the FMC() class.")

    API_CONFIG_VERSION = "api/fmc_config/v1"
    API_PLATFORM_VERSION = "api/fmc_platform/v1"
    VERIFY_CERT = False
    MAX_PAGING_REQUESTS = 2000
    TOO_MANY_CONNECTIONS_TIMEOUT = 30
    FMC_MAX_PAYLOAD = 2048000

    def __init__(
        self,
        host="192.168.45.45",
        username="admin",
        password="Admin123",
        domain=None,
        autodeploy=True,
        file_logging=None,
        logging_level="INFO",
        debug=False,
        limit=1000,
        timeout=5,
        wait_time=15,
        api_key=None,
        uuid=None,
    ):
        """
        Instantiate some variables prior to calling the __enter__() method.

        :param host (str): Hostname/IP of FMC (Default is 192.168.45.45)
        :param username (str): Admin for FMC (Default is admin)
        :param password (str): Admin Password (Default is Admin123)
        :param domain (str): UUID of domain (Default is None which implies the Global domain)
        :param autodeploy (bool): Deploy changes to affected FMC manage devices. (Default is True)
        :param file_logging (str): The filename (and optional path) of the output file if a file logger is required,
        None if no file logger is required. (Default is None)
        :param logging_level (str): The desired logging level. (Default is INFO)
        :param debug (bool): True to enable debug logging. (Default is False)
        :param limit (int): Sets up max data to gather per "page". (Default is 1000)
        :param timeout (int):  Maximum seconds to establish connection (Default is 5)
        :return: None
        """
        self.debug = debug
        if self.debug:
            logging_level = "DEBUG"
        root_logger = logging.getLogger("")
        if logging_level.upper() == "DEBUG":
            root_logger.setLevel(logging.DEBUG)
        if logging_level.upper() == "INFO":
            root_logger.setLevel(logging.INFO)
        if logging_level.upper() == "WARNING":
            root_logger.setLevel(logging.WARNING)
        if logging_level.upper() == "ERROR":
            root_logger.setLevel(logging.ERROR)
        if logging_level.upper() == "CRITICAL":
            root_logger.setLevel(logging.CRITICAL)

        if file_logging:
            print(
                f'Logging is enabled and set to {logging_level}.  Look for file "{file_logging}" for output.'
            )
            formatter = logging.Formatter(
                "%(asctime)s - %(levelname)s:%(filename)s:%(lineno)s - %(message)s",
                "%Y/%m/%d-%H:%M:%S",
            )
            file_logger = RotatingFileHandler(
                file_logging, maxBytes=1024000, backupCount=10, mode="w"
            )
            file_logger.setFormatter(formatter)
            root_logger.addHandler(file_logger)

        logging.debug("In the FMC __init__() class method.")

        self.host = host
        self.username = username
        self.password = password
        self.domain = domain
        self.autodeploy = autodeploy
        self.limit = limit
        self.timeout = timeout
        self.vdbVersion = None
        self.sruVersion = None
        self.serverVersion = None
        self.geoVersion = None
        self.configuration_url = None
        self.platform_url = None
        self.page_counter = None
        self.more_items = []
        self.error_response = None
        self.wait_time = wait_time
        self.api_key = api_key
        self.uuid = uuid

    def __enter__(self):
        """
        Get a token from the FMC as well as the Global UUID.  With this information set up the base_url variable.

        :return: self
        """
        logging.debug("In the FMC __enter__() class method.")
        if self.api_key is None:
            self.mytoken = Token(
                host=self.host,
                username=self.username,
                password=self.password,
                domain=self.domain,
                verify_cert=self.VERIFY_CERT,
                timeout=self.timeout,
            )
            if not self.mytoken.access_token:
                logging.info("User authentication failed.")
                exit(1)
            self.uuid = self.mytoken.uuid

        else:
            if self.uuid is None:
                logging.error("If using an API_KEY, you must provide a UUID")
                exit(1)

        self.build_urls()
        version = ServerVersion(fmc=self)
        version.get()
        self.serverVersion = version.serverVersion
        logging.info(f"This FMC's version is {self.serverVersion}")
        return self

    def __exit__(self, *args):
        """
        If autodeploy == True, push changes to FMC upon exit of "with" contract.

        :param args:
        :return: None
        """
        logging.debug("In the FMC __exit__() class method.")

        if self.autodeploy:
            tmp = DeploymentRequests(fmc=self)
            tmp.post()
        else:
            logging.info(
                "Auto deploy changes set to False.  Use the Deploy button in FMC to push changes to FTDs."
            )

    def build_urls(self):
        """
        Build configuration_url and platform_url variables.

        :return: None
        """
        logging.debug("In the FMC build_urls() class method.")
        logging.info("Building base to URLs.")
        self.configuration_url = (
            f"https://{self.host}/{self.API_CONFIG_VERSION}/domain/{self.uuid}"
        )
        self.platform_url = f"https://{self.host}/{self.API_PLATFORM_VERSION}"

    def send_to_api(
        self, method="", url="", headers="", json_data=None, more_items=None
    ):
        """
        Send API call to FMC.

        :param method (str): GET, POST, PUT, or DELETE
        :param url (str): URL for API call.
        :param headers (str):  String of header variables.
        :param json_data (str):  JSON formatted string as payload. (Default is None)
        :param more_items (str):  Used for paging in query.
        :return: JSON response from FMC
        """
        logging.debug("In the FMC send_to_api() class method.")

        if not more_items:
            self.more_items = []
            self.page_counter = 0
        if self.api_key is not None:
            headers = {
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.api_key}",
            }
        else:
            if headers == "":
                # These values for headers works for most API requests.
                headers = {
                    "Content-Type": "application/json",
                    "X-auth-access-token": self.mytoken.get_token(),
                }

        status_code = 429
        response = None
        json_response = None
        logging.debug(
            f"Being sent to FMC's API:\n\tHEADERS={headers}\n\tURL={url}\n\tMETHOD={method}\n\t"
            f"MORE_ITEMS={more_items}\n\tJSON_DATA={json_data}"
        )
        try:
            while status_code == 429:
                if method == "get":
                    response = requests.get(
                        url,
                        headers=headers,
                        verify=self.VERIFY_CERT,
                        timeout=self.timeout,
                    )
                elif method == "post":
                    response = requests.post(
                        url,
                        json=json_data,
                        headers=headers,
                        verify=self.VERIFY_CERT,
                        timeout=self.timeout,
                    )
                elif method == "put":
                    response = requests.put(
                        url,
                        json=json_data,
                        headers=headers,
                        verify=self.VERIFY_CERT,
                        timeout=self.timeout,
                    )
                elif method == "delete":
                    response = requests.delete(
                        url,
                        headers=headers,
                        verify=self.VERIFY_CERT,
                        timeout=self.timeout,
                    )
                else:
                    logging.error("No request method given.  Returning nothing.")
                    return
                if self.debug:
                    debug_msg = ["Response from FMC's API:"]
                    for response_var in dir(response):
                        if "__" not in response_var:
                            tmp_var = getattr(response, response_var)
                            debug_msg.append(f"\n\t{response_var}={tmp_var}")
                    logging.debug("".join(debug_msg))

                status_code = response.status_code
                if status_code == 429:
                    logging.warning(
                        f"Too many connections to the FMC.  Waiting {self.TOO_MANY_CONNECTIONS_TIMEOUT} "
                        f"seconds and trying again."
                    )
                    time.sleep(self.TOO_MANY_CONNECTIONS_TIMEOUT)
                if status_code == 401:
                    if self.api_key is not None:
                        logging.warning("Token has expired. Trying to refresh.")
                        self.mytoken.access_token = None
                        self.mytoken.access_token = self.mytoken.get_token()
                        headers = {
                            "Content-Type": "application/json",
                            "X-auth-access-token": self.mytoken.access_token,
                        }
                        status_code = 429
                    else:
                        logging.warning(
                            "Received HTTP Code 401 from FMC. Please check that your API key is valid and has the correct permissions in CDO/FMC"
                        )
                if status_code == 422:
                    logging.warning(
                        "Either:\n\t1. Payload too large.  FMC can only handle a payload of "
                        f"{self.FMC_MAX_PAYLOAD} bytes.\n\t2.The payload contains an unprocessable or "
                        f"unreadable entity such as a invalid attribut name or incorrect JSON syntax "
                    )
            json_response = json.loads(response.text)
            if status_code > 301 or "error" in json_response:
                response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            logging.error(f"Error in POST operation --> {str(err)}")
            logging.error(f"json_response -->\t{json_response}")
            self.error_response = json_response
            if response:
                response.close()
            return None
        if response:
            response.close()
        try:
            if (
                "next" in json_response["paging"]
                and self.page_counter <= self.MAX_PAGING_REQUESTS
            ):
                self.more_items += json_response["items"]
                logging.debug(
                    f"Paging:  Offset:{json_response['paging']['offset']}, "
                    f"Limit:{json_response['paging']['limit']}, "
                    f"Count:{json_response['paging']['count']}, "
                    f"Gathered_Items:{len(self.more_items)}."
                )
                self.page_counter += 1
                return self.send_to_api(
                    method=method,
                    url=json_response["paging"]["next"][0],
                    json_data=json_data,
                    more_items=self.more_items,
                )
            else:
                json_response["items"] = self.more_items + json_response["items"]
                self.more_items = []
                return json_response
        except KeyError:
            # Used only when the response only has "one page" of results.
            return json_response


class Token(object):
    """The token is the validation object used with the FMC."""

    logging.debug("In the Token class.")

    MAX_REFRESHES = 3
    TOKEN_LIFETIME = 60 * 30
    TOKEN_REFRESH_TIME = int(
        TOKEN_LIFETIME * 0.95
    )  # Refresh token at 95% refresh time.
    API_PLATFORM_VERSION = "api/fmc_platform/v1"

    def __init__(
        self,
        host="192.168.45.45",
        username="admin",
        password="Admin123",
        domain=None,
        verify_cert=False,
        timeout=5,
    ):
        """
        Initialize variables used in the Token class.

        :param host (str):  FMC hostname/IP (Default is 192.168.45.45)
        :param username (str): FMC Admin user (Default is admin)
        :param password (str): FMC user's password (Default is Admin123)
        :param domain (str):  UUID of domain.  Default is None which implies Global domain.
        :param verify_cert (bool):  Validate cert  (Default is False)
        :param timeout (int):  Maximum seconds to establish connection (Default is 5)
        :return: None
        """
        logging.debug("In the Token __init__() class method.")

        self.__host = host
        self.__username = username
        self.__password = password
        self.__domain = domain
        self.uuid = None
        self.verify_cert = verify_cert
        self.timeout = timeout
        self.token_refreshes = 0
        self.access_token = None
        self.refresh_token = None
        self.token_creation_time = None
        self.generate_tokens()

    def generate_tokens(self):
        """
        Create new or refresh expired tokens.

        :return: None
        """
        logging.debug("In the Token generate_tokens() class method.")

        if self.token_refreshes <= self.MAX_REFRESHES and self.access_token is not None:
            headers = {
                "Content-Type": "application/json",
                "X-auth-access-token": self.access_token,
                "X-auth-refresh-token": self.refresh_token,
            }
            url = f"https://{self.__host}/{self.API_PLATFORM_VERSION}/auth/refreshtoken"
            logging.info(
                f"Refreshing tokens, {self.token_refreshes} out of {self.MAX_REFRESHES} refreshes, "
                f"from {url}."
            )
            response = requests.post(
                url, headers=headers, verify=self.verify_cert, timeout=self.timeout
            )
            logging.debug(
                "Response from refreshtoken() post:\n"
                f"\turl: {url}\n"
                f"\theaders: {headers}\n"
                f"\tresponse: {response}"
            )
            self.token_refreshes += 1
        else:
            self.token_refreshes = 0
            self.token_creation_time = (
                datetime.datetime.now()
            )  # Can't trust that your clock is in sync with FMC's.
            headers = {"Content-Type": "application/json"}
            url = (
                f"https://{self.__host}/{self.API_PLATFORM_VERSION}/auth/generatetoken"
            )
            logging.info(f"Requesting new tokens from {url}.")
            response = requests.post(
                url,
                headers=headers,
                auth=requests.auth.HTTPBasicAuth(self.__username, self.__password),
                verify=self.verify_cert,
                timeout=self.timeout,
            )
            logging.debug(
                "Response from generatetoken() post:\n"
                f"\turl: {url}\n"
                f"\theaders: {headers}\n"
                f"\tresponse: {response}"
            )
        self.access_token = response.headers.get("X-auth-access-token")
        self.refresh_token = response.headers.get("X-auth-refresh-token")
        self.uuid = response.headers.get("DOMAIN_UUID")
        if self.access_token:
            all_domain = json.loads(response.headers.get("DOMAINS"))
            if self.__domain is not None:
                for domain in all_domain:
                    if "global/" + self.__domain.lower() == domain["name"].lower():
                        logging.info(f"Domain set to {domain['name']}")
                        self.uuid = domain["uuid"]
                    else:
                        logging.info(
                            "Domain name entered not found in FMC, falling back to Global"
                        )

    def get_token(self):
        """
        Check validity of current token.  If needed make a new or refresh.  Then return access_token.

        :return self.access_token
        """
        logging.debug("In the Token get_token() class method.")
        if (
            datetime.datetime.now()
            > (
                self.token_creation_time
                + datetime.timedelta(seconds=self.TOKEN_REFRESH_TIME)
            )
            or self.access_token == None
        ):
            logging.info("Token expired.  Generating a new token.")
            self.token_refreshes = 0
            self.access_token = None
            self.refresh_token = None
            self.generate_tokens()

        return self.access_token



"""IPSec Settings Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .ftds2svpns import FTDS2SVPNs
from fmcapi.api_objects.object_services.ikev1ipsecproposals import IKEv1IpsecProposals
from fmcapi.api_objects.object_services.ikev2ipsecproposals import IKEv2IpsecProposals
import logging


class IPSecSettings(APIClassTemplate):
    """The IPSecSettings Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "tfcPackets",
        "validateIncomingIcmpErrorMessage",
        "enableSaStrengthEnforcement",
        "perfectForwardSecrecy",
        "lifetimeSeconds",
        "doNotFragmentPolicy",
        "lifetimeKilobytes",
        "cryptoMapType",
        "ikeV2Mode",
        "enableRRI",
        "ikeV2IpsecProposal",
        "ikeV1IpsecProposal",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    FIRST_SUPPORTED_FMC_VERSION = "6.3"
    PREFIX_URL = "/policy/ftds2svpns"
    REQUIRED_FOR_POST = ["vpn_id"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize IPSecSettings object.

        Set self.type to "IPSecSettings" and parse the kwargs.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for IPSecSettings class.")
        self.parse_kwargs(**kwargs)
        self.type = "IPSecSettings"

    def vpn_policy(self, pol_name):
        """
        Associate a VPN Policy.

        :param pol_name: (str) Name of VPN Policy.
        :return: None
        """
        logging.debug("In vpn_policy() for IPSecSettings class.")
        ftd_s2s = FTDS2SVPNs(fmc=self.fmc)
        ftd_s2s.get(name=pol_name)
        if "id" in ftd_s2s.__dict__:
            self.vpn_id = ftd_s2s.id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.vpn_id}/ipsecsettings"
            self.vpn_added_to_url = True
        else:
            logging.warning(
                f'FTD S2S VPN Policy "{pol_name}" not found.  Cannot set up IPSecSettings for FTDS2SVPNs Policy.'
            )

    def ipsec_policy(self, pol_name, version=1, action="add"):
        """
        Associate IPSec Policy.

        :param pol_name: (str) Name of IPSec Policy.
        :param version: (int) Version number.
        :param action: (str) 'add', 'remove', or 'clear'.
        """
        logging.debug("In ipsec_policy() for IPSecSettings class.")
        pol1 = None
        if version == 1:
            pol1 = IKEv1IpsecProposals(fmc=self.fmc)
        elif version == 2:
            pol1 = IKEv2IpsecProposals(fmc=self.fmc)
        else:
            logging.warning("Invalid version type specified.  Must be 1 or 2")
        pol1.get(name=pol_name)
        new_pol = None

        if action == "add":
            if "id" in pol1.__dict__ and version == 1:
                new_pol = {"id": pol1.id, "name": pol1.name, "type": pol1.type}
            elif "id" in pol1.__dict__ and version == 2:
                new_pol = {"id": pol1.id, "name": pol1.name, "type": pol1.type}

            if new_pol is None:
                logging.warning(
                    f'IKEv"{version}"IpsecProposal "{pol_name}" not found.  Cannot set up IPSecSettings Policy.'
                )
            else:
                if "version" == 1:
                    logging.info(f'Adding "{pol_name}" to ikeV1IpsecProposal.')
                    if "ikeV1IpsecProposal" in self.__dict__:
                        self.ikeV1IpsecProposal.append(new_pol)
                    else:
                        self.ikeV1IpsecProposal = [new_pol]
                elif "version" == 2:
                    logging.info(f'Adding "{pol_name}" to ikeV2IpsecProposal.')
                    if "ikeV2IpsecProposal" in self.__dict__:
                        self.ikeV2IpsecProposal.append(new_pol)
                    else:
                        self.ikeV2IpsecProposal = [new_pol]

        elif action == "remove":
            if "version" == 1:
                if "ikeV1IpsecProposal" in self.__dict__:
                    self.ikeV1IpsecProposal = list(
                        filter(lambda i: i["name"] != pol_name, self.ikeV1IpsecProposal)
                    )
                else:
                    logging.warning(
                        "ikeV1IpsecProposal has no members.  Cannot remove policy."
                    )
            if "version" == 2:
                if "ikeV2IpsecProposal" in self.__dict__:
                    self.ikeV2IpsecProposal = list(
                        filter(lambda i: i["name"] != pol_name, self.ikeV2IpsecProposal)
                    )
                else:
                    logging.warning(
                        "ikeV2IpsecProposal has no members.  Cannot remove policy."
                    )

        elif action == "clear":
            if "version" == 1:
                if "ikeV1IpsecProposal" in self.__dict__:
                    del self.ikeV1IpsecProposal
            if "version" == 2:
                if "ikeV2IpsecProposal" in self.__dict__:
                    del self.ikeV2IpsecProposal



import logging
import fmcapi


def test__dynamic_objects_mappings(fmc):
    logging.info("Testing Dynamic Objects Mappings")

    # obj1 = fmcapi.DynamicObjectMappings(fmc)
    # obj1.handle_mappings(action="add", value=["10.0.0.4"], name="Sales")
    # obj1.post()

    obj2 = fmcapi.DynamicObjectMappings(fmc=fmc, id="0050568E-B2F8-0ed3-0000-523986013212")
    #obj2.mappings = ["70.0.0.0/24"]
    logging.info(f"Dynamic object mapping {obj2.get()}")










import logging
import fmcapi


def test__application_filter(fmc):
    logging.info("Testing ApplicationFilter class.")
    obj1 = fmcapi.ApplicationFilters(fmc=fmc)
    logging.info("All ApplicationFilters -- >")
    result = obj1.get()
    logging.info(result)
    # There are no Application Filters by default so there is no items in the list.
    if "items" in result:
        logging.info(f"Total items: {len(result['items'])}")
    del obj1

    logging.info("Testing ApplicationFilter class done.\n")



import logging
import fmcapi


def test__audit_records(fmc):
    logging.info("Testing AuditRecords() class.")
    tmp = fmcapi.AuditRecords(fmc=fmc)
    logging.info(tmp.get())
    logging.info("Testing AuditRecords() class done.\n")



import logging
import fmcapi
import time


def test__acp_rule(fmc):
    logging.info(
        "In preparation for testing ACPRule methods, set up some known objects in the FMC."
    )

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    # Build an IP host object
    iphost1 = fmcapi.Hosts(fmc=fmc, name="_iphost1", value="7.7.7.7")
    iphost1.post()
    # Build an IP Network object
    ipnet1 = fmcapi.Networks(fmc=fmc, name="_ipnet1", value="1.2.3.0/24")
    ipnet1.post()
    # Build an IP range object
    iprange1 = fmcapi.Ranges(fmc=fmc, name="_iprange1", value="6.6.6.6-7.7.7.7")
    iprange1.post()
    # Build a Network Group object
    ipnet2 = fmcapi.Networks(fmc=fmc, name="_ipnet2", value="5.5.5.0/24")
    ipnet2.post()
    time.sleep(1)
    # Build an FQDNS object
    fqdns1 = fmcapi.FQDNS(fmc=fmc, name="_fqdns1", value="www.cisco.com")
    fqdns1.post()

    obj1 = fmcapi.NetworkGroups(fmc=fmc, name="_fmcapi_test_networkgroup")
    obj1.named_networks(action="add", name=ipnet2.name)
    obj1.unnamed_networks(action="add", value="4.4.4.4/32")
    obj1.post()
    # Build a URL object
    url1 = fmcapi.URLs(fmc=fmc, name="_url1", url="asdf.org")
    url1.post()
    url1.get()
    # lists = [{"type": url1.type, "id": url1.id, "name": url1.name}]
    # Build a VLAN Tag object
    vlantag1 = fmcapi.VlanTags(
        fmc=fmc, name="_vlantag1", data={"startTag": "888", "endTag": "999"}
    )
    vlantag1.post()
    # Build a Port object
    pport1 = fmcapi.ProtocolPortObjects(
        fmc=fmc, name="_pport1", port="9090", protocol="UDP"
    )
    pport1.post()
    # Build a Port Group Object
    obj10 = fmcapi.ProtocolPortObjects(
        fmc=fmc, name="_porttcp1", port="8443", protocol="TCP"
    )
    obj10.post()
    obj11 = fmcapi.ProtocolPortObjects(
        fmc=fmc, name="_portudp1", port="161", protocol="UDP"
    )
    obj11.post()
    obj12 = fmcapi.ProtocolPortObjects(
        fmc=fmc, name="_portrangetcp1", port="0-1023", protocol="TCP"
    )
    obj12.post()
    obj2 = fmcapi.PortObjectGroups(fmc=fmc, name="_fmcapi_test_portobjectgroup")
    obj2.named_ports(action="add", name=obj10.name)
    obj2.named_ports(action="add", name=obj11.name)
    obj2.named_ports(action="add", name=obj12.name)
    obj2.post()
    # Build a Security Zone object
    sz1 = fmcapi.SecurityZones(fmc=fmc, name="_sz1", interfaceMode="ROUTED")
    sz1.post()
    # Build an ACP Object
    acp1 = fmcapi.AccessPolicies(fmc=fmc, name=namer)
    acp1.post()
    # Get a file_policy
    # fp = fmcapi.FilePolicies(fmc=fmc1, name='daxm_test')
    time.sleep(1)
    logging.info("Setup of objects for ACPRule test done.\n")

    logging.info(
        "Test ACPRule.  Try to test all features of all methods of the ACPRule class."
    )
    acprule1 = fmcapi.AccessRules(fmc=fmc, acp_name=acp1.name)
    acprule1.name = namer
    acprule1.action = "ALLOW"
    acprule1.enabled = False
    acprule1.sendEventsToFMC = True
    acprule1.logFiles = False
    acprule1.logBegin = True
    acprule1.logEnd = True
    acprule1.variable_set(action="set", name="Default-Set")
    acprule1.source_zone(action="add", name=sz1.name)
    acprule1.destination_zone(action="add", name=sz1.name)
    acprule1.intrusion_policy(action="set", name="Security Over Connectivity")
    acprule1.vlan_tags(action="add", name=vlantag1.name)
    acprule1.source_port(action="add", name=pport1.name)
    acprule1.destination_port(action="add", name=pport1.name)
    acprule1.destination_port(action="add", name=obj2.name)
    acprule1.source_network(action="add", name=iphost1.name)
    acprule1.source_network(action="add", name=obj1.name)
    acprule1.source_network(action="add", name=iprange1.name)
    acprule1.destination_network(action="add", name=ipnet1.name)
    acprule1.destination_network(action="add", name=iprange1.name)
    acprule1.destination_network(action="add", name=fqdns1.name)
    acprule1.urls_info(action="add", name=url1.name)

    # To set a comment on an ACP rule, you use the new_comments function which sets "newComments" in the API POST call
    # To read comments, you read the "commentHistoryList" from the API GET call
    # Just another one of those strange quirks of the FMC API
    acprule1.new_comments(action="add", value="comment-1")
    acprule1.post()

    logging.info("Test ACPRule done.\n")

    logging.info("Cleanup of testing ACPRule methods.")
    acprule1.delete()
    time.sleep(1)
    acp1.delete()
    iphost1.delete()
    ipnet1.delete()
    iprange1.delete()
    fqdns1.delete()
    obj1.delete()
    ipnet2.delete()
    url1.delete()
    vlantag1.delete()
    pport1.delete()
    sz1.delete()
    obj2.delete()
    obj10.delete()
    obj11.delete()
    obj12.delete()
    logging.info("Cleanup of objects for ACPRule test done.\n")



"""Etherchannel Interfaces Classes."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .devicerecords import DeviceRecords
from fmcapi.api_objects.object_services.securityzones import SecurityZones
from fmcapi.api_objects.device_services.physicalinterfaces import PhysicalInterfaces
import logging


class EtherchannelInterfaces(APIClassTemplate):
    """The EtherchannelInterfaces Interface Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "mode",
        "enabled",
        "MTU",
        "managementOnly",
        "ipAddress",
        "selectedInterfaces",
        "etherChannelId",
        "lacpMode",
        "maxActivePhysicalInterface",
        "minActivePhysicalInterface",
        "hardware",
        "erspanFlowId",
        "erspanSourceIP",
        "loadBalancing",
        "macLearn",
        "ifname",
        "securityZone",
        "arpConfig",
        "ipv4",
        "ipv6",
        "macTable",
        "enableAntiSpoofing",
        "fragmentReassembly",
        "enableDNSLookup",
        "activeMACAddress",
        "standbyMACAddress",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + ["device_name"]
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\-\/\. ]"""
    PREFIX_URL = "/devices/devicerecords"
    URL_SUFFIX = None
    REQUIRED_FOR_POST = ["etherChannelId", "mode", "MTU"]
    REQUIRED_FOR_PUT = ["id", "device_id"]
    VALID_FOR_IPV4 = ["static", "dhcp", "pppoe"]
    VALID_FOR_MODE = ["INLINE", "PASSIVE", "TAP", "ERSPAN", "NONE"]
    VALID_FOR_LACP_MODE = ["ACTIVE", "PASSIVE", "ON"]
    VALID_FOR_MTU = range(64, 9085)
    VALID_FOR_LOAD_BALANCING = (
        []
    )  # Not sure what to put here but it was an unresolved variable later in this code.
    VALID_FOR_HARDWARE_SPEED = [
        "AUTO",
        "TEN",
        "HUNDRED",
        "THOUSAND",
        "TEN_THOUSAND",
        "FORTY_THOUSAND",
        "LAKH",
    ]
    VALID_FOR_HARDWARE_DUPLEX = ["AUTO", "FULL", "HALF"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize EtherchannelInterfaces object.

        Set self.type to "RedundantInterface" and parse the kwargs.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for RedundantInterfaces class.")
        self.parse_kwargs(**kwargs)
        self.type = "RedundantInterface"

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for EtherchannelInterfaces class.")
        if "device_name" in kwargs:
            self.device(device_name=kwargs["device_name"])
        if "ipv4" in kwargs:
            if list(kwargs["ipv4"].keys())[0] in self.VALID_FOR_IPV4:
                self.ipv4 = kwargs["ipv4"]
            else:
                logging.warning(f"Method {kwargs['ipv4']} is not a valid ipv4 type.")
        if "mode" in kwargs:
            if kwargs["mode"] in self.VALID_FOR_MODE:
                self.mode = kwargs["mode"]
            else:
                logging.warning(f"Mode {kwargs['mode']} is not a valid mode.")
        if "MTU" in kwargs:
            if kwargs["MTU"] in self.VALID_FOR_MTU:
                self.MTU = kwargs["MTU"]
            else:
                logging.warning(f"MTU {kwargs['MTU']} should be in the range 64-9000.")
                self.MTU = 1500
        if "lacpMode" in kwargs:
            if kwargs["lacpMode"] in self.VALID_FOR_LACP_MODE:
                self.lacpMode = kwargs["lacpMode"]
            else:
                logging.warning(f"LACP Mode {kwargs['lacpMode']} is not a valid mode.")
        if "loadBalancing" in kwargs:
            if kwargs["loadBalancing"] in self.VALID_FOR_LOAD_BALANCING:
                self.loadBalancing = kwargs["loadBalancing"]
            else:
                logging.warning(
                    f"Load balancing method {kwargs['loadBalancing']} is not a valid method."
                )
        """
        if "hardware" in kwargs:
            if kwargs["hardware"] in self.VALID_FOR_HARDWARE:
                self.hardware = kwargs["hardware"]
            else:
                logging.warning(
                    f"Hardware method {kwargs['hardware']} is not a valid method.")
        """

    def device(self, device_name):
        """
        Associate device to EtherChannel.

        :param device_name: (str) Name of device.
        :return: None
        """
        logging.debug("In device() for EtherchannelInterfaces class.")
        device1 = DeviceRecords(fmc=self.fmc)
        device1.get(name=device_name)
        if "id" in device1.__dict__:
            self.device_id = device1.id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.device_id}/etherchannelinterfaces"
            self.device_added_to_url = True
        else:
            logging.warning(
                f"Device {device_name} not found.  Cannot set up device for EtherchannelInterfaces."
            )

    def sz(self, name):
        """
        Assign Security Zone to EtherChannel.

        :param name: (str) Name of Security Zone.
        :return: None
        """
        logging.debug("In sz() for EtherchannelInterfaces class.")
        sz = SecurityZones(fmc=self.fmc)
        sz.get(name=name)
        if "id" in sz.__dict__:
            new_zone = {"name": sz.name, "id": sz.id, "type": sz.type}
            self.securityZone = new_zone
        else:
            logging.warning(
                f'Security Zone, "{name}", not found.  Cannot add to RedundantInterfaces.'
            )

    def static(self, ipv4addr, ipv4mask):
        """
        Assign static IP to this EtherChannel.

        :param ipv4addr: (str) x.x.x.x
        :param ipv4mask: (str) bitmask
        :return: None
        """
        logging.debug("In static() for EtherchannelInterfaces class.")
        self.ipv4 = {"static": {"address": ipv4addr, "netmask": ipv4mask}}

    def dhcp(self, enableDefault=True, routeMetric=1):
        """
        Configure this EtherChannel with DHCP for addressing.

        :param enableDefault: (bool) Accept, or not, a default route via DHCP.
        :param routeMetric: (int) Set route metric.
        :return: None
        """
        logging.debug("In dhcp() for EtherchannelInterfaces class.")
        self.ipv4 = {
            "dhcp": {
                "enableDefaultRouteDHCP": enableDefault,
                "dhcpRouteMetric": routeMetric,
            }
        }

    def p_interfaces(self, p_interfaces, device_name):
        """
        Define which physical interface on which device is a part of this EtherChannel.

        :param p_interfaces: (str) Name of physical interface.
        :param device_name: (str) Name of device with that interface.
        :return: None
        """
        logging.debug("In p_interfaces() for EtherchannelInterfaces class.")
        list1 = []
        for p_intf in p_interfaces:
            intf1 = PhysicalInterfaces(fmc=self.fmc)
            intf1.get(name=p_intf, device_name=device_name)
            if "id" in intf1.__dict__:
                list1.append({"name": intf1.name, "id": intf1.id, "type": intf1.type})
            else:
                logging.warning(
                    f'PhysicalInterface, "{intf1.name}", not found.  Cannot add to EtherchannelInterfaces.'
                )
        self.selectedInterfaces = list1

    def hardware(self, speed, duplex):
        """
        Define hardware characteristics.

        :param speed: (str) Speed of interface.
        :param duplex: (str) AUTO FULL or HALF.
        :return: None
        """
        # There are probably some incompatibilities that need to be accounted for
        logging.debug("In hardware() for EtherchannelInterfaces class.")
        if (
            speed in self.VALID_FOR_HARDWARE_SPEED
            and duplex in self.VALID_FOR_HARDWARE_DUPLEX
        ):
            self.hardware = {"duplex": duplex, "speed": speed}
        else:
            logging.warning(f"Speed {speed} or Duplex {duplex} is not a valid mode.")



import logging
from .hit_counts import test__hitcounts
from .url_category import test__url_category
from .ports import test__ports
from .upgrades import test__upgrades
from .manualnat import test__manualnat
from .autonat import test__autonat
from .port_object_group import test__port_object_group
from .acprule import test__acp_rule
from .acp import test__access_control_policy
from .intrusion_policy import test__intrusion_policy
from .interfaces_subinterfaces import test__subinterfaces
from .interfaces_etherchannel import test__etherchannel_interfaces
from .interfaces_redundant import test__redundant_interfaces
from .interfaces_bridge_group import test__bridge_group_interfaces
from .interfaces_physical import test__phys_interfaces
from .wait_for_task import wait_for_task
from .device_with_task import test__device_with_task
from .sla_monitor import test__slamonitor
from .interface_group import test__interface_group
from .security_zone import test__security_zone
from .protocol_port import test__protocol_port
from .vlan_tag import test__vlan_tag
from .url import test__url
from .ikev2 import test__ikev2
from .ikev1 import test__ikev1
from .icmpv6 import test__icmpv6
from .icmpv4 import test__icmpv4
from .geolocations import test__geolocations
from .acls_extended import test__extended_acls
from .ip_range import test__ip_range
from .ip_network import test__ip_network
from .ip_host import test__ip_host
from .objects_get_query_filters import test__objects_get_query_filters
from .variable_set import test__variable_set
from .server_version import test__fmc_version
from .ip_addresses import test__ip_addresses
from .network_group import test__network_group
from .url_group import test__url_group
from .vlan_group_tag import test__vlan_group_tag
from .dns_servers_group import test__dns_servers_group
from .continent import test__continent
from .file_policies import test__filepolicies
from .country import test__country
from .certificate_enrollment import test__cert_enrollment
from .application_category import test__application_category
from .application_productivity import test__application_productivity
from .application_filter import test__application_filter
from .application_risk import test__application_risk
from .application import test__application
from .application_tag import test__application_tag
from .application_type import test__application_type

from .audit_records import test__audit_records
from .deployable_devices import test__deployable_devices
from .deployment_requests import test__deployment_requests
from .devicegrouprecords import test__devicegrouprecords
from .ftddevicehapairs import test__ftddevicehapairs
from .failoverinterfacemacaddressconfigs import test__failoverinterfacemacaddressconfigs
from .monitored_interface import test__monitoredinterfaces
from .devicerecords import test__devicerecords
from .staticroutes import test__staticroutes
from .ipv4staticroutes import test__ipv4staticroutes
from .ipv6staticroutes import test__ipv6staticroutes
from .prefilter import test__prefilter_policy
from .prefilter_rule import test__prefiler_rule
from .s2s_vpn import test__ftds2svpns
from .search_globalsearch import test__globalsearch
from .search_object import test__object
from .search_policy import test__policy
from .backup import test__backup
from .usage import test__usage
from .dynamic_objects import test__dynamic_objects
from .dynamic_object_mappings import test__dynamic_objects_mappings
from .terminateravpnsessions import test__terminateravpnsessions
from .grouppolicies import test__grouppolicies
from .ravpn import test__ravpn
from .connectionprofiles import test__connectionprofiles
from .dynamicaccesspolicies import test__dynamicaccesspolicies
from .timeranges import test__timeranges

logging.debug("In the unit-tests __init__.py file.")

__all__ = [
    "test__audit_records",
    "test__deployment_requests",
    "test__deployable_devices",
    "test__devicegrouprecords",
    "test__ftddevicehapairs",
    "test__failoverinterfacemacaddressconfigs",
    "test__monitoredinterfaces",
    "test__devicerecords",
    "test__staticroutes",
    "test__ipv4staticroutes",
    "test__ipv6staticroutes",
    "test__bridge_group_interfaces",
    "test__url_category",
    "test__application_type",
    "test__application_tag",
    "test__application",
    "test__application_risk",
    "test__application_filter",
    "test__application_productivity",
    "test__application_category",
    "test__cert_enrollment",
    "test__country",
    "test__filepolicies",
    "test__continent",
    "test__dns_servers_group",
    "test__vlan_group_tag",
    "test__url_group",
    "test__network_group",
    "test__ip_addresses",
    "test__fmc_version",
    "test__variable_set",
    "test__ip_host",
    "test__objects_get_query_filters",
    "test__ip_network",
    "test__ip_range",
    "test__extended_acls",
    "test__geolocations",
    "test__icmpv6",
    "test__icmpv4",
    "test__ikev2",
    "test__ikev1",
    "test__vlan_tag",
    "test__url",
    "test__protocol_port",
    "test__security_zone",
    "test__interface_group",
    "test__slamonitor",
    "test__device_with_task",
    "wait_for_task",
    "test__phys_interfaces",
    "test__redundant_interfaces",
    "test__etherchannel_interfaces",
    "test__subinterfaces",
    "test__ports",
    "test__upgrades",
    "test__manualnat",
    "test__autonat",
    "test__port_object_group",
    "test__acp_rule",
    "test__access_control_policy",
    "test__intrusion_policy",
    "test__prefilter_policy",
    "test__hitcounts",
    "test__prefiler_rule",
    "test__ftds2svpns",
    "test__globalsearch",
    "test__object",
    "test__policy",
    "test__backup",
    "test__usage",
    "test__dynamic_objects",
    "test__dynamic_objects_mappings",
    "test__terminateravpnsessions",
    "test__grouppolicies",
    "test__ravpn",
    "test__connectionprofiles",
    "test__dynamicaccesspolicies",
    "test__timeranges",
]



"""Auto NAT Rules class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .ftdnatpolicies import FTDNatPolicies
from fmcapi.api_objects.object_services.networkaddresses import NetworkAddresses
from fmcapi.api_objects.object_services.interfaceobjects import InterfaceObjects
from fmcapi.api_objects.object_services.networkgroups import NetworkGroups
import logging


class AutoNatRules(APIClassTemplate):
    """The AutoNatRules Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "originalNetwork",
        "translatedNetwork",
        "interfaceInTranslatedNetwork",
        "natType",
        "interfaceIpv6",
        "fallThrough",
        "dns",
        "routeLookup",
        "noProxyArp",
        "netToNet",
        "sourceInterface",
        "destinationInterface",
        "originalPort",
        "translatedPort",
        "serviceProtocol",
        "patOptions",
        "description",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    PREFIX_URL = "/policy/ftdnatpolicies"
    REQUIRED_FOR_POST = ["nat_id"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize AutoNatRules object.

        Set self.type to "FTDAutoNatRule" and parse the kwargs.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for AutoNatRules class.")
        self.parse_kwargs(**kwargs)
        self.type = "FTDAutoNatRule"

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for AutoNatRules class.")
        if ("translatedNetwork" in kwargs) and (
            "interfaceInTranslatedNetwork" in kwargs
        ):
            logging.warning(
                "Cannot have both a translatedNetwork and interfaceInTranslatedNetwork"
            )
        elif "translatedNetwork" in kwargs:
            self.translatedNetwork = kwargs["translatedNetwork"]
        elif "interfaceInTranslatedNetwork" in kwargs:
            self.interfaceInTranslatedNetwork = kwargs["interfaceInTranslatedNetwork"]

    def nat_policy(self, name):
        """
        Associate NAT Policy with this rule.

        :param name: (str) Name of NAT Policy.
        :return: None
        """
        logging.debug("In nat_policy() for AutoNatRules class.")
        ftd_nat = FTDNatPolicies(fmc=self.fmc)
        ftd_nat.get(name=name)
        if "id" in ftd_nat.__dict__:
            self.nat_id = ftd_nat.id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.nat_id}/autonatrules"
            self.nat_added_to_url = True
        else:
            logging.warning(
                f"FTD NAT Policy {name} not found.  Cannot set up AutoNatRule for NAT Policy."
            )

    def original_network(self, name):
        """
        Associate Network with this rule.

        :param name: (str) Name of Network.
        :return: None
        """
        logging.debug("In original_network() for AutoNatRules class.")
        ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
        items = ipaddresses_json.get("items", [])
        new_net = None
        for item in items:
            if item["name"] == name:
                new_net = {"id": item["id"], "type": item["type"]}
                break
        if new_net is None:
            logging.warning(
                f'Network "{name}" is not found in FMC.  Cannot add to originalNetwork.'
            )
        else:
            self.originalNetwork = new_net
            logging.info(f'Adding "{name}" to sourceNetworks for this AutoNatRule.')

    def translated_network(self, name):
        """
        Associate Network to this rule.

        :param name: (str) Name of Network.
        :return: None
        """
        # Auto Nat rules can't use network group objects
        logging.debug("In translated_network() for AutoNatRules class.")
        ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
        items = ipaddresses_json.get("items", [])
        new_net = None
        for item in items:
            if item["name"] == name:
                new_net = {"id": item["id"], "type": item["type"]}
                break
        if new_net is None:
            logging.warning(
                f'Network "{name}" is not found in FMC.  Cannot add to translatedNetwork.'
            )
        else:
            self.translatedNetwork = new_net
            logging.info(
                f'Adding "{name}" to destinationNetworks for this AutoNatRule.'
            )

    def source_intf(self, name):
        """
        Associate interface to this rule.

        :param name: (str) Name of interface.
        :return: None
        """
        logging.debug("In source_intf() for AutoNatRules class.")
        intf_obj = InterfaceObjects(fmc=self.fmc).get()
        items = intf_obj.get("items", [])
        new_intf = None
        for item in items:
            if item["name"] == name:
                new_intf = {"id": item["id"], "type": item["type"]}
                break
        if new_intf is None:
            logging.warning(
                f'Interface Object "{name}" is not found in FMC.  Cannot add to sourceInterface.'
            )
        else:
            if new_intf["type"] == "InterfaceGroup" and len(new_intf.interfaces) > 1:
                logging.warning(
                    f'Interface Object "{name}" contains more than one physical interface. Cannot add to '
                    f"sourceInterface."
                )
            else:
                self.sourceInterface = new_intf
                logging.info(f'Interface Object "{name}" added to NAT Policy.')

    def destination_intf(self, name):
        """
        Associate interface to this rule.

        :param name: (str) Name of interface.
        :return: None
        """
        logging.debug("In destination_intf() for AutoNatRules class.")
        intf_obj = InterfaceObjects(fmc=self.fmc).get()
        items = intf_obj.get("items", [])
        new_intf = None
        for item in items:
            if item["name"] == name:
                new_intf = {"id": item["id"], "type": item["type"]}
                break
        if new_intf is None:
            logging.warning(
                f'Interface Object "{name}" is not found in FMC.  Cannot add to destinationInterface.'
            )
        else:
            if new_intf["type"] == "InterfaceGroup" and len(new_intf.interfaces) > 1:
                logging.warning(
                    f'Interface Object "{name}" contains more than one physical interface. Cannot add to '
                    f"destinationInterface."
                )
            else:
                self.destinationInterface = new_intf
                logging.info(f'Interface Object "{name}" added to NAT Policy.')

    def identity_nat(self, name):
        """
        Associate Network to this rule.

        :param name: (str) Name of Network.
        :return: None
        """
        logging.debug("In identity_nat() for AutoNatRules class.")
        ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
        items = ipaddresses_json.get("items", [])
        new_net = None
        for item in items:
            if item["name"] == name:
                new_net = {"id": item["id"], "type": item["type"]}
                break
        if new_net is None:
            logging.warning(
                f'Network "{name}" is not found in FMC.  Cannot add to this AutoNatRule.'
            )
        else:
            self.natType = "STATIC"
            self.originalNetwork = new_net
            self.translatedNetwork = new_net
            logging.info(f'Adding "{name}" to AutoNatRule.')

    def patPool(self, name, options={}):
        """
        Associate a PAT Pool with this rule.

        :param name: (str) Name of PAT Pool.
        :param options: (dict) Dictionary of options.
        :return: None
        """
        # Network Group Object permitted for patPool
        ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
        networkgroup_json = NetworkGroups(fmc=self.fmc).get()
        items = ipaddresses_json.get("items", []) + networkgroup_json.get("items", [])
        new_net = None
        for item in items:
            if item["name"] == name:
                new_net = {"name": item["name"], "id": item["id"], "type": item["type"]}
                break
        if new_net is None:
            logging.warning(
                f'Network "{name}" is not found in FMC.  Cannot add to patPool.'
            )
        else:
            self.natType = "DYNAMIC"
            self.patOptions = {"patPoolAddress": new_net}
            self.patOptions["interfacePat"] = (
                options["interfacePat"] if "interfacePat" in options.keys() else False
            )
            self.patOptions["includeReserve"] = (
                options["includeReserve"]
                if "includeReserve" in options.keys()
                else False
            )
            self.patOptions["roundRobin"] = (
                options["roundRobin"] if "roundRobin" in options.keys() else True
            )
            self.patOptions["extendedPat"] = (
                options["extendedPat"] if "extendedPat" in options.keys() else False
            )
            self.patOptions["flatPortRange"] = (
                options["flatPortRange"] if "flatPortRange" in options.keys() else False
            )
            logging.info(f'Adding "{name}" to patPool for this AutoNatRule.')



"""IKE v1 Policies Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class IKEv1Policies(APIClassTemplate):
    """The IKEv1Policies Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "encryption",
        "hash",
        "priority",
        "diffieHellmanGroup",
        "authenticationMethod",
        "lifetimeInSeconds",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/object/ikev1policies"
    REQUIRED_FOR_POST = [
        "name",
        "encryption",
        "hash",
        "diffieHellmanGroup",
        "lifetimeInSeconds",
        "authenticationMethod",
    ]
    VALID_FOR_ENCRYPTION = ["DES", "3DES", "AES-128", "AES-192", "AES-256"]
    VALID_FOR_HASH = ["MD5", "SHA"]
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""
    FIRST_SUPPORTED_FMC_VERSION = "6.3"

    def __init__(self, fmc, **kwargs):
        """
        Initialize IKEv1Policies object.

        Set self.type to "IKEv1PolicyObject" and parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for IKEv1Policies class.")
        self.parse_kwargs(**kwargs)
        self.type = "Ikev1PolicyObject"

    def format_data(self, filter_query=""):
        """
        Gather all the data in preparation for sending to API in JSON format.

        :param filter_query: (str) 'all' or 'kwargs'
        :return: (dict) json_data
        """
        json_data = super().format_data()
        logging.debug("In format_data() for IKEv1Policies class.")
        if "encryption" in self.__dict__:
            if self.encryption in self.VALID_FOR_ENCRYPTION:
                json_data["encryption"] = self.encryption
            else:
                logging.warning(f'encryption "{self.encryption}" not a valid type.')
        if "hash" in self.__dict__:
            if self.hash in self.VALID_FOR_HASH:
                json_data["hash"] = self.hash
            else:
                logging.warning(f'hash "{self.hash}" not a valid type.')
        return json_data



"""Protocol Port Objects Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class ProtocolPortObjects(APIClassTemplate):
    """The ProtocolPortObjects in the FMC."""

    VALID_JSON_DATA = ["id", "name", "description", "port", "protocol", "type"]
    VALID_GET_FILTERS = [
        "unusedOnly",
        "nameOrValue",
    ]  # unusedOnly:Bool, nameOrValue:String
    VALID_FOR_KWARGS = VALID_JSON_DATA + VALID_GET_FILTERS + []
    URL_SUFFIX = "/object/protocolportobjects"
    REQUIRED_FOR_POST = ["name", "protocol"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize ProtocolPortObjects object.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for ProtocolPortObjects class.")
        self.parse_kwargs(**kwargs)



import logging
import fmcapi
import time
import json

def test__connectionprofiles(fmc):
    logging.info("Testing ConnectionProfiles class.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    connection_profiles = fmcapi.ConnectionProfiles(fmc=fmc, container_uuid="40A6B737-FDDC-0ed3-0000-227633610030")
    connection_profiles.get(expanded=True)

    # Example to change group policy in an existing connection profile
    # for item in conn_profs.get('items'):
    #     conn_profile_id = item.get('id')
    #     profile = fmcapi.ConnectionProfiles(fmc=fmc, container_uuid="40A6B737-FDDC-0ed3-0000-227633610030")
    #     profile.id = conn_profile_id
    #     conn_prof = profile.get(expanded=True)
    #     logging.info(f"current groupPolicy: {conn_prof.get('groupPolicy')}")
    #     lock_out_grp_pol = {
    #         'name': 'LOCKOUT_GRP_POL',
    #         'id': '40A6B737-FDDC-0ed3-0000-227637209390',
    #         'type': 'GroupPolicy'
    #     }
    #     profile.groupPolicy = lock_out_grp_pol
    #     profile.put()

    logging.info("Testing ConnectionProfiles class done.\n")


"""IPv6 Static Routes Classes."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .devicerecords import DeviceRecords
from fmcapi.api_objects.object_services.networkaddresses import NetworkAddresses
from fmcapi.api_objects.object_services.hosts import Hosts
from fmcapi.api_objects.object_services.networkgroups import NetworkGroups
import logging


class IPv6StaticRoutes(APIClassTemplate):
    """The IPv6StaticRoute Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "interfaceName",
        "selectedNetworks",
        "gateway",
        "routeTracking",
        "metricValue",
        "isTunneled",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + ["device_name"]
    PREFIX_URL = "/devices/devicerecords"
    URL_SUFFIX = None
    REQUIRED_FOR_POST = ["interfaceName", "selectedNetworks", "gateway"]
    REQUIRED_FOR_PUT = ["id", "device_id"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize IPv6StaticRoutes object.

        Set self.type to "IP6StaticRoute" and parse the kwargs.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for IPv6StaticRoute class.")
        self.type = "IPv6StaticRoute"
        self.parse_kwargs(**kwargs)

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for IPv6StaticRoute class.")
        if "device_name" in kwargs:
            self.device(device_name=kwargs["device_name"])

    def device(self, device_name):
        """
        Associate device to this route.

        :param device_name: (str) Name of device.
        :return: None
        """
        logging.debug("In device() for IPv6StaticRoute class.")
        device1 = DeviceRecords(fmc=self.fmc)
        device1.get(name=device_name)
        if "id" in device1.__dict__:
            self.device_id = device1.id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.device_id}/routing/ipv6staticroutes"
            self.device_added_to_url = True
        else:
            logging.warning(
                f"Device {device_name} not found.  Cannot set up device for IPv6StaticRoute."
            )

    def networks(self, action, networks):
        """
        Set of networks associated with this route.

        :param action: (str) 'add', 'remove', or 'clear'
        :param networks: (list)  List of IP Hosts, IP Networks, and/or Network Groups.
        :return: None
        """
        logging.info("In networks() for IPv6StaticRoute class.")
        if action == "add":
            # Valid objects are IPHost, IPNetwork and NetworkGroup.
            # Create a dictionary to contain all three object type.
            ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
            networkgroup_json = NetworkGroups(fmc=self.fmc).get()
            items = ipaddresses_json.get("items", []) + networkgroup_json.get(
                "items", []
            )
            for network in networks:
                # Find the matching object name in the dictionary if it exists
                net1 = list(filter(lambda i: i["name"] == network, items))
                if len(net1) > 0:
                    if "selectedNetworks" in self.__dict__:
                        # Check to see if network already exists
                        exists = list(
                            filter(
                                lambda i: i["id"] == net1[0]["id"],
                                self.selectedNetworks,
                            )
                        )
                        if "id" in exists:
                            logging.warning(
                                f'Network "{network}" already exists in selectedNetworks.'
                            )
                        else:
                            self.selectedNetworks.append(
                                {
                                    "type": net1[0]["type"],
                                    "id": net1[0]["id"],
                                    "name": net1[0]["name"],
                                }
                            )
                    else:
                        self.selectedNetworks = [
                            {
                                "type": net1[0]["type"],
                                "id": net1[0]["id"],
                                "name": net1[0]["name"],
                            }
                        ]
                else:
                    logging.warning(
                        f'Network "{network}" not found.  Cannot set up device for IPv6StaticRoute.'
                    )
        elif action == "remove":
            ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
            networkgroup_json = NetworkGroups(fmc=self.fmc).get()
            items = ipaddresses_json.get("items", []) + networkgroup_json.get(
                "items", []
            )
            for network in networks:
                net1 = list(filter(lambda i: i["name"] == network, items))
                if len(net1) > 0:
                    if "selectedNetworks" in self.__dict__:
                        self.selectedNetworks = list(
                            filter(
                                lambda i: i["id"] != net1[0]["id"],
                                self.selectedNetworks,
                            )
                        )
                    else:
                        logging.warning(
                            "No selectedNetworks found for this Device's IPv6StaticRoute."
                        )
                else:
                    logging.warning(
                        f'Network "{network}" not found.  Cannot set up device for IPv6StaticRoute.'
                    )
        elif action == "clear":
            if "selectedNetworks" in self.__dict__:
                del self.selectedNetworks
                logging.info(
                    "All selectedNetworks removed from this IPv6StaticRoute object."
                )

    def gw(self, name):
        """
        Gateway for this route.

        :param name: (str) Name of object that is the gateway address.
        :return: None
        """
        logging.info("In gw() for IPv6StaticRoute class.")
        gw1 = Hosts(fmc=self.fmc)
        gw1.get(name=name)
        if "id" in gw1.__dict__:
            self.gateway = {
                "object": {"type": gw1.type, "id": gw1.id, "name": gw1.name}
            }
        else:
            logging.warning(
                f'Network "{name}" not found.  Cannot set up device for IPv6StaticRoute.'
            )



"""Policy Assignments Services Classes."""

import logging
from .policyassignments import PolicyAssignments

logging.debug("In the policy_assignment_services __init__.py file.")

__all__ = ["PolicyAssignments"]



"""Group Policies Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class GroupPolicies(APIClassTemplate):
    """The GroupPolicies Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "generalSettings",  # Dict
        "anyConnectSettings",  # Dict
        "advancedSettings",  # Dict
        "enableSSLProtocol",  # Bool
        "enableIPsecIKEv2Protocol",  # Bool
    ]
    VALID_GET_FILTERS = [
        "unusedOnly",  # Bool
        "nameOrValue",  # String
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + VALID_GET_FILTERS + []
    URL_SUFFIX = "/object/grouppolicies"
    REQUIRED_FOR_PUT = [
        "id",
        "name",
        "generalSettings",  # Dict
        "anyConnectSettings",  # Dict
        "advancedSettings",  # Dict
        "enableSSLProtocol",  # Bool
        "enableIPsecIKEv2Protocol",  # Bool
    ]

    def __init__(self, fmc, **kwargs):
        """
        Initialize GroupPolicies object.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for GroupPolicies class.")
        self.type = "GroupPolicy"
        self.parse_kwargs(**kwargs)



"""FailoverInterfaceMACAddressConfigs class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .ftddevicehapairs import FTDDeviceHAPairs
from fmcapi.api_objects.device_services.physicalinterfaces import PhysicalInterfaces
import logging


class FailoverInterfaceMACAddressConfigs(APIClassTemplate):
    """The FailoverInterfaceMACAddressConfigs Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "physicalInterface",
        "failoverActiveMac",
        "failoverStandbyMac",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + ["ha_name"]
    PREFIX_URL = "/devicehapairs/ftddevicehapairs"
    REQUIRED_FOR_POST = ["physicalInterface", "failoverActiveMac", "failoverStandbyMac"]
    REQUIRED_FOR_PUT = ["id"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize FailoverInterfaceMACAddressConfigs object.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for FailoverInterfaceMACAddressConfigs class.")
        self.parse_kwargs(**kwargs)

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for FailoverInterfaceMACAddressConfigs class.")
        if "ha_name" in kwargs:
            self.device_ha(ha_name=kwargs["ha_name"])

    def device_ha(self, ha_name):
        """
        Add deviceha_id to URL.

        :param ha_name: (str) Name of Device HA.
        :return: None
        """
        logging.debug("In device_ha() for FailoverInterfaceMACAddressConfigs class.")
        deviceha1 = FTDDeviceHAPairs(fmc=self.fmc, name=ha_name)
        deviceha1.get()
        if "id" in deviceha1.__dict__:
            self.deviceha_id = deviceha1.id
            self.URL = (
                f"{self.fmc.configuration_url}{self.PREFIX_URL}/"
                f"{self.deviceha_id}/failoverinterfacemacaddressconfigs"
            )
            self.deviceha_added_to_url = True
        else:
            logging.warning(
                f"Device HA {ha_name} not found. "
                f"Cannot set up device for FailoverInterfaceMACAddressConfigs."
            )

    def p_interface(self, name, device_name):
        """
        Physical interface of device used for HA.

        :param name: (str) Name of interface.
        :param device_name (str) Name of device.
        :return: None
        """
        logging.debug("In p_interface() for FailoverInterfaceMACAddressConfigs class.")
        intf1 = PhysicalInterfaces(fmc=self.fmc)
        intf1.get(name=name, device_name=device_name)
        if "id" in intf1.__dict__:
            self.physicalInterface = {
                "name": intf1.name,
                "id": intf1.id,
                "type": intf1.type,
            }
        else:
            logging.warning(
                f'PhysicalInterface, "{name}", not found.  '
                f"Cannot add to FailoverInterfaceMACAddressConfigs."
            )

    def edit(self, name, ha_name):
        """
        Edit existing device HA and change physical interface params.

        :param name: (str)
        :param ha_name: (str)
        :return: None
        """
        logging.debug("In edit() for FailoverInterfaceMACAddressConfigs class.")
        deviceha1 = FTDDeviceHAPairs(fmc=self.fmc, name=ha_name)
        deviceha1.get()
        obj1 = FailoverInterfaceMACAddressConfigs(fmc=self.fmc)
        obj1.device_ha(ha_name=ha_name)
        failovermac_json = obj1.get()
        items = failovermac_json.get("items", [])
        found = False
        for item in items:
            if item["physicalInterface"]["name"] == name:
                found = True
                self.id = item["id"]
                self.name = item["physicalInterface"]["name"]
                self.failoverActiveMac = item["failoverActiveMac"]
                self.failoverStandbyMac = item["failoverStandbyMac"]
                self.deviceha_id = deviceha1.id
                self.URL = (
                    f"{self.fmc.configuration_url}{self.PREFIX_URL}/"
                    f"{self.deviceha_id}/failoverinterfacemacaddressconfigs"
                )
                break
        if found is False:
            logging.warning(
                f'PhysicalInterface, "{name}", not found.  Cannot add to FailoverInterfaceMACAddressConfigs.'
            )



import logging
import fmcapi


def test__variable_set(fmc):
    logging.info("Test VariableSet. Can only GET VariableSet objects.")

    obj1 = fmcapi.VariableSets(fmc=fmc)
    obj1.get(name="Default-Set")
    logging.info("VariableSet -->")
    logging.info(obj1.format_data())

    logging.info("Test VariableSet done.\n")



"""Endpoints Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .ftds2svpns import FTDS2SVPNs
from fmcapi.api_objects.object_services.fqdns import FQDNS
from fmcapi.api_objects.object_services.hosts import Hosts
from fmcapi.api_objects.object_services.networks import Networks
from fmcapi.api_objects.object_services.networkgroups import NetworkGroups
from fmcapi.api_objects.device_ha_pair_services.ftddevicehapairs import FTDDeviceHAPairs
from fmcapi.api_objects.device_services.devicerecords import DeviceRecords
from fmcapi.api_objects.device_services.etherchannelinterfaces import (
    EtherchannelInterfaces,
)
from fmcapi.api_objects.device_services.physicalinterfaces import PhysicalInterfaces
from fmcapi.api_objects.device_services.redundantinterfaces import RedundantInterfaces
from fmcapi.api_objects.device_services.subinterfaces import SubInterfaces
import logging


class Endpoints(APIClassTemplate):
    """The Endpoints Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "device",
        "interface",
        "nattedInterfaceAddress",
        "protectedNetworks",
        "ipv6InterfaceAddress",
        "connectionType",
        "peerType",
        "extranet",
        "extranetInfo",
        "description",
        "version",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    FIRST_SUPPORTED_FMC_VERSION = "6.3"
    VALID_FOR_POINT_TO_POINT = ["PEER"]
    VALID_FOR_HUB_AND_SPOKE = ["HUB", "SPOKE"]
    PREFIX_URL = "/policy/ftds2svpns"
    REQUIRED_FOR_POST = ["vpn_id"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize Endpoints object.

        Set self.type to "Endpoint" and parse the kwargs.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for Endpoints class.")
        self.parse_kwargs(**kwargs)
        self.type = "EndPoint"

    def vpn_policy(self, pol_name):
        """
        Associate a VPN Policy.

        :param pol_name: (str) Name of VPN Policy.
        :return: None
        """
        logging.debug("In vpn_policy() for Endpoints class.")
        ftd_s2s = FTDS2SVPNs(fmc=self.fmc)
        ftd_s2s.get(name=pol_name)
        if "id" in ftd_s2s.__dict__:
            self.vpn_id = ftd_s2s.id
            self.URL = (
                f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.vpn_id}/endpoints"
            )
            self.vpn_added_to_url = True
            self.topology_type = ftd_s2s.topologyType
        else:
            logging.warning(
                f'FTD S2S VPN Policy "{pol_name}" not found.  '
                f"Cannot set up Endpoints for FTDS2SVPNs Policy."
            )

    def endpoint(self, action, device_name):
        """
        Associate an endpoint.

        :param action: (str) 'add', 'remove', or 'clear'
        :param device_name: (str) Name of device.
        """
        logging.debug("In endpoint() for Endpoints class.")
        device_json = DeviceRecords(fmc=self.fmc).get()
        device_ha_json = FTDDeviceHAPairs(fmc=self.fmc).get()
        items = device_json.get("items", []) + device_ha_json.get("items", [])
        new_device = None

        if action == "add":
            for item in items:
                if item["name"] == device_name:
                    new_device = {
                        "name": item["name"],
                        "id": item["id"],
                        "type": item["type"],
                    }
                    break
            if new_device is None:
                logging.warning(
                    f'Device/DeviceHA "{device_name}" is not found in FMC.  Cannot add to Endpoints.'
                )
            else:
                if "device" in self.__dict__:
                    self.device.append(new_device)
                    logging.info(f'Adding "{device_name}" to Endpoints.')
                else:
                    self.device = new_device
        elif action == "remove":
            if "device" in self.__dict__:
                self.device = list(
                    filter(lambda i: i["name"] != device_name, self.device)
                )
            else:
                logging.warning("Endpoints has no members.  Cannot remove device.")
        elif action == "clear":
            if "device" in self.__dict__:
                del self.device

    def vpn_interface(self, device_name, ifname):
        """
        Associate an interface.

        :param device_name: (str) Name of device.
        :param ifname: (str) Name of interface.
        """
        logging.debug("In vpn_interface() for Endpoints class.")
        ether_json = EtherchannelInterfaces(fmc=self.fmc, device_name=device_name).get()
        phys_json = PhysicalInterfaces(fmc=self.fmc, device_name=device_name).get()
        redund_json = RedundantInterfaces(fmc=self.fmc, device_name=device_name).get()
        subintf_json = SubInterfaces(fmc=self.fmc, device_name=device_name).get()
        items = (
            ether_json.get("items", [])
            + phys_json.get("items", [])
            + redund_json.get("items", [])
            + subintf_json.get("items", [])
        )
        new_intf = None
        for item in items:
            if item["ifname"] == ifname:
                new_intf = {"id": item["id"], "type": item["type"]}
                break
        if new_intf is None:
            logging.warning(
                f'Interface "{ifname}" is not found in FMC.  Cannot add to interface.'
            )
        else:
            self.interface = new_intf
            logging.info(f'Interface "{ifname}" added.')

    def encryption_domain(self, action, names=[]):
        """
        Associate Encryption.

        :param action: (str) 'add', 'remove', or 'clear'.
        :param names: (list) List of Encryption names.
        """
        logging.debug("In endpoint() for Endpoints class.")
        fqdns_json = FQDNS(fmc=self.fmc).get()
        host_json = Hosts(fmc=self.fmc).get()
        net_json = Networks(fmc=self.fmc).get()
        netg_json = NetworkGroups(fmc=self.fmc).get()
        items = (
            fqdns_json.get("items", [])
            + host_json.get("items", [])
            + net_json.get("items", [])
            + netg_json.get("items", [])
        )
        new_network = None

        if action == "add":
            for name in names:
                for item in items:
                    if item["name"] == name:
                        new_network = {"id": item["id"], "type": item["type"]}
                        break
                if new_network is None:
                    logging.warning(
                        f'FQDNS/Host/Network/Network Group"{name}" is not found in FMC.'
                        f"  Cannot add to protectedNetworks."
                    )
                else:
                    if "protectedNetworks" in self.__dict__:
                        self.protectedNetworks["networks"].append(new_network)
                        logging.info(f'Appending "{name}" to protectedNetworks.')
                    else:
                        self.protectedNetworks = {"networks": [new_network]}
                        logging.info(f'Adding "{name}" to protectedNetworks.')
        elif action == "remove":
            if "protectedNetworks" in self.__dict__:
                for name in names:
                    self.protectedNetworks = list(
                        filter(lambda i: i["name"] != name, self.protectedNetworks)
                    )
            else:
                logging.warning(
                    "protectedNetworks has no members.  Cannot remove network."
                )
        elif action == "clear":
            if "protectedNetworks" in self.__dict__:
                del self.protectedNetworks



import logging
import fmcapi
import time

def test__timeranges(fmc):
    logging.info("Testing TimeRanges class.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    time_range = fmcapi.TimeRanges(fmc=fmc)
    time_range.name = namer
    time_range.effectiveStartDateTime = "1979-01-01T00:00"
    time_range.effectiveEndDateTime = "1979-01-01T00:01"
    time_range.post()
    time_range.get()
    time_range.effectiveStartDateTime = "1979-01-02T00:00"
    time_range.effectiveEndDateTime = "1979-01-02T00:01"
    time_range.put()
    time_range.delete()

    logging.info("Testing TimeRanges class done.\n")


import logging
import fmcapi
import time


def test__access_control_policy(fmc):
    logging.info("Test AccessControlPolicy.  Post, get, put, delete ACP Objects.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    obj1 = fmcapi.AccessPolicies(fmc=fmc)
    obj1.name = namer
    obj1.post()
    time.sleep(1)
    del obj1
    obj1 = fmcapi.AccessPolicies(fmc=fmc, name=namer)
    obj1.get()
    obj1.name = "asdfasdf"
    obj1.put()
    time.sleep(1)
    obj1.delete()
    logging.info("Test AccessControlPolicy done.\n")



"""Usage Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from fmcapi.api_objects.helper_functions import *
import logging


class Usage(APIClassTemplate):
    """The Usage Object in the FMC."""

    VALID_JSON_DATA = []
    VALID_GET_FILTERS = [
        "uuid",
        "type",
    ]  # uuid:String, type:String[Network,Port,VLAN,URL]
    REQUIRED_GET_FILTERS = ["uuid", "type"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + VALID_GET_FILTERS + []
    URL_SUFFIX = "/object/operational/usage"

    def __init__(self, fmc, **kwargs):
        """
        Initialize Usage object.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for Usage class.")
        self.parse_kwargs(**kwargs)



import logging
import fmcapi
import time


def test__ftddevicehapairs(fmc):
    logging.info(
        'Test FTDDeviceHAPairs. After an HA Pair is created, all API calls to "devicerecords" objects should '
        "be directed at the currently active device not the ha pair"
    )
    failover1 = fmcapi.PhysicalInterfaces(fmc=fmc)
    failover1.get(device_name="PrimaryName", name="GigabitEthernet0/6")
    stateful1 = fmcapi.PhysicalInterfaces(fmc=fmc)
    stateful1.get(device_name="PrimaryName", name="GigabitEthernet0/7")
    obj0 = fmcapi.DeviceHAPairs(fmc=fmc)
    obj1 = fmcapi.FTDDeviceHAPairs(fmc=fmc)
    obj1.primary(name="PrimaryName")
    obj1.secondary(name="SecondaryName")
    obj1.name = "HaName"
    # failover interface subnetMask must be in x.x.x.x format"
    obj1.ftdHABootstrap = {
        "isEncryptionEnabled": "true",
        "encKeyGenerationScheme": "CUSTOM",
        "sharedKey": "cisco123",
        "useSameLinkForFailovers": False,
        "lanFailover": {
            "useIPv6Address": False,
            "subnetMask": "255.255.255.252",
            "interfaceObject": {
                "type": "PhysicalInterface",
                "name": failover1.name,
                "id": failover1.id,
            },
            "standbyIP": "192.168.1.2",
            "logicalName": "HA-FAILOVER",
            "activeIP": "192.168.1.1",
        },
        "statefulFailover": {
            "useIPv6Address": False,
            "subnetMask": "255.255.255.252",
            "interfaceObject": {
                "type": "PhysicalInterface",
                "name": stateful1.name,
                "id": stateful1.id,
            },
            "standbyIP": "192.168.1.6",
            "logicalName": "HA-STATEFUL",
            "activeIP": "192.168.1.5",
        },
    }
    # response = ha_pair.post()
    # wait_for_task(response["metadata"]["task"], 30)
    logging.info("Device HA-->")
    logging.info(obj1.format_data())
    logging.info("\n")
    obj1.post()
    time.sleep(300)
    del obj1

    obj1 = fmcapi.FTDDeviceHAPairs(fmc=fmc, name="HaName")
    obj1.switch_ha()
    logging.info("Device HA Switch-->")
    logging.info(obj1.format_data())
    logging.info("\n")
    response = obj1.put()
    logging.info(response)
    time.sleep(20)
    del obj1

    obj1 = fmcapi.FTDDeviceHAPairs(fmc=fmc, name="HaName")
    obj1.break_ha()
    logging.info("Device HA Break-->")
    logging.info(obj1.format_data())
    logging.info("\n")
    response = obj1.put()
    logging.info(response)

    time.sleep(300)
    del obj1
    obj1 = fmcapi.FTDDeviceHAPairs(fmc=fmc)
    obj1.get(name="FTDv-HA2")

    #  Deleting the HAPair object will delete the HA configuration AND remove the devices from the FMC
    obj1.delete()



from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class Object(APIClassTemplate):
    """
    The Object Object in the FMC.
    """

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "category",
        "links",
        "self",
        "metadata",
        "domain",
    ]
    VALID_FOR_KWARGS = ["filter"]
    VALID_CHARACTERS_FOR_NAME = """[ .\w\d_\-]"""
    URL_SUFFIX = "/search/object"
    REQUIRED_FOR_GET = ["filter"]
    FILTER_BY_NAME = True
    FIRST_SUPPORTED_FMC_VERSION = "7.0"

    def __init__(self, fmc, **kwargs):
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for Object class.")
        self.parse_kwargs(**kwargs)
        self.URL = f"{self.URL}?filter={self.filter}"
        pass

    def format_data(self):
        json_data = super().format_data()
        logging.debug("In format_data() for Object class.")
        return json_data



import logging
import fmcapi


def test__monitoredinterfaces(fmc):
    logging.info("Test MonitoredInterfaces. get, put MonitoredInterfaces Objects")

    obj0 = fmcapi.DeviceHAMonitoredInterfaces(fmc=fmc, ha_name="HaName")
    obj1 = fmcapi.MonitoredInterfaces(fmc=fmc, ha_name="HaName")
    # Interface logical name (ifname)
    obj1.get(name="OUTSIDE1")
    obj1.monitorForFailures = True
    obj1.ipv4(ipv4addr="10.254.0.4", ipv4mask=29, ipv4standbyaddr="10.254.0.3")
    logging.info("MonitoredInterfaces PUT-->")
    logging.info(obj1.format_data())

    logging.info(obj1.put())



import logging
import fmcapi
import time


def test__url_group(fmc):
    logging.info("Testing URLGroup class.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    url1 = fmcapi.URLs(fmc=fmc, name="_url1", url="example.org")
    url1.post()
    url2 = fmcapi.URLs(fmc=fmc, name="_url2", url="example.net")
    url2.post()
    url3 = fmcapi.URLs(fmc=fmc, name="_url3", url="example.com")
    url3.post()
    time.sleep(1)
    obj1 = fmcapi.URLGroups(fmc=fmc, name=namer)
    obj1.named_urls(action="add", name=url1.name)
    obj1.named_urls(action="add", name=url1.name)
    obj1.named_urls(action="clear")
    obj1.named_urls(action="add", name=url2.name)
    obj1.named_urls(action="add", name=url3.name)
    obj1.named_urls(action="add", name=url1.name)
    obj1.named_urls(action="remove", name=url3.name)
    obj1.post()
    time.sleep(1)
    del obj1
    obj1 = fmcapi.URLGroups(fmc=fmc, name=namer)
    obj1.get()
    obj1.unnamed_urls(action="add", value="daxm.net")
    obj1.unnamed_urls(action="add", value="daxm.com")
    obj1.unnamed_urls(action="clear")
    obj1.unnamed_urls(action="add", value="daxm.org")
    obj1.unnamed_urls(action="add", value="daxm.net")
    obj1.unnamed_urls(action="add", value="daxm.lan")
    obj1.unnamed_urls(action="remove", value="daxm.org")
    obj1.put()
    time.sleep(1)
    obj1.delete()
    url1.delete()
    url2.delete()
    url3.delete()

    logging.info("Testing URLGroup class done.\n")



"""Continents Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class Continents(APIClassTemplate):
    """The Continents Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "countries"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/object/continents"
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""

    def __init__(self, fmc, **kwargs):
        """
        Initialize Continents object.

        Set self.type to "Continent" and parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for Continents class.")
        self.parse_kwargs(**kwargs)
        self.type = "Continent"

    def post(self):
        """POST method for API for Continents not supported."""
        logging.info("POST method for API for Continents not supported.")
        pass

    def put(self):
        """PUT method for API for Continents not supported."""
        logging.info("PUT method for API for Continents not supported.")
        pass

    def delete(self):
        """DELETE method for API for Continents not supported."""
        logging.info("DELETE method for API for Continents not supported.")
        pass



"""Default Actions Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .accesspolicies import AccessPolicies
import logging


class DefaultActions(APIClassTemplate):
    """The DefaultActions Object in the FMC."""

    VALID_JSON_DATA = []
    VALID_FOR_KWARGS = VALID_JSON_DATA + [
        "acp_id",
        "acp_name",
        "device_id",
        "device_name",
        "fetchZeroHitCount",
        "limit",
        "action",
    ]
    PREFIX_URL = "/policy/accesspolicies"
    REQUIRED_FOR_PUT = ["acp_id", "id", "action"]
    REQUIRED_FOR_GET = ["acp_id"]
    VALID_ACTION = ["BLOCK", "TRUST", "PERMIT", "NETWORK_DISCOVERY"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize DefaultActions object.

        Set self.type to "AccessPolicyDefaultAction", parse the kwargs, and set up the self.URL.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        logging.debug("In __init__() for DefaultActions class.")
        super().__init__(fmc, **kwargs)
        self.parse_kwargs(**kwargs)
        self.type = "AccessPolicyDefaultAction"
        self.URL = f"{self.URL}{self.URL_SUFFIX}"

    def format_data(self, filter_query=""):
        """
        Gather all the data in preparation for sending to API in JSON format.

        :param filter_query: (str) 'all' or 'kwargs'
        :return: (dict) json_data
        """
        json_data = super().format_data()
        logging.debug("In format_data() for DefaultActions class.")
        if "action" in self.__dict__:
            if self.action in self.VALID_ACTION:
                json_data["action"] = self.action
            else:
                logging.warning(f"action variable must be one of: {self.VALID_ACTION}.")
        return json_data

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for DefaultActions class.")
        if "acp_id" in kwargs:
            self.acp(acp_id=kwargs["acp_id"])
        if "acp_name" in kwargs:
            self.acp(name=kwargs["acp_name"])
        if "device_id" in kwargs:
            self.device(id=kwargs["device_id"])
        if "device_name" in kwargs:
            self.device(name=kwargs["device_name"])
        if "action" in kwargs:
            if kwargs["action"] in self.VALID_ACTION:
                self.action = kwargs["action"]
            else:
                logging.warning(f"action variable must be one of: {self.VALID_ACTION}.")
            self.action = kwargs["action"]

    def acp(self, name="", acp_id=""):
        """
        Associate an AccessPolicies object with this DefaultAction object.

        :param name: (str)  Name of ACP.
        :param id: (str) ID of ACP.
        :return: None
        """
        # either name or id of the ACP should be given
        logging.debug("In acp() for DefaultActions class.")
        if acp_id != "":
            self.acp_id = acp_id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.acp_id}"
            self.acp_added_to_url = True
        elif name != "":
            acp1 = AccessPolicies(fmc=self.fmc)
            acp1.get(name=name)
            if "id" in acp1.__dict__:
                self.acp_id = acp1.id
                self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.acp_id}/defaultactions"
                self.acp_added_to_url = True
            else:
                logging.warning(
                    f'Access Control Policy "{name}" not found.  Cannot configure acp for DefaultActions.'
                )
        else:
            logging.error("No accessPolicy name or id was provided.")

    def post(self):
        """POST method for DefaultActions not supported."""
        logging.info("API POST method for DefaultActions not supported.")
        pass

    def delete(self, **kwargs):
        """DELETE method for DefaultActions not supported."""
        logging.info("API DELETE method for DefaultActions not supported.")
        pass



"""The fmcapi __init__.py file is called whenever someone imports the package into their program."""

import logging
from .fmc import FMC
from .api_objects import *

logging.debug("In the fmcapi __init__.py file.")


def __authorship__():
    """
    Authorship and gratitude for contributors.

    This python module was created by Dax Mickelson along with LOTs of help from Ryan Malloy and Neil Patel.
    Thank you to the github community members who have also pitched in, especially Mark Sullivan and his team.  Feel
    free to send comments/suggestions/improvements.  Either by email: dmickels@cisco.com or more importantly via a pull
    request from the github repository: https://github.com/daxm/fmcapi.
    """
    logging.debug(__authorship__.__doc__)


__authorship__()



import logging
import fmcapi
import time


def test__protocol_port(fmc):
    logging.info("Test ProtocolPort.  Post, get, put, delete Port Objects.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    obj1 = fmcapi.ProtocolPortObjects(fmc=fmc)
    obj1.name = namer
    obj1.port = "1234"
    obj1.protocol = "TCP"
    obj1.post()
    time.sleep(1)
    del obj1
    obj1 = fmcapi.ProtocolPortObjects(fmc=fmc, name=namer)
    obj1.get()
    obj1.port = "5678"
    obj1.put()
    time.sleep(1)
    obj1.delete()
    logging.info("Test ProtocolPort done.\n")
